#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.pending_api.py

    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     26 Oct 2020, (2:26 PM)

    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved

           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:

           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.

           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""

import json
import os
import time
import tornado.web
import tornado.log
import tornado.routing

from unmanic import config
from unmanic.libs.uiserver import UnmanicDataQueues
from unmanic.webserver.api_v1.base_api_handler import BaseApiHandler

from unmanic.libs import task, common
from unmanic.webserver.helpers import pending_tasks


class ApiPendingHandler(BaseApiHandler):
    name = None
    config = None
    params = None
    unmanic_data_queues = None

    routes = [
        {
            "supported_methods": ["POST"],
            "call_method":       "create_task_from_path",
            "path_pattern":      r"/api/v1/pending/add/",
        },
        {
            "supported_methods": ["POST"],
            "call_method":       "manage_pending_tasks_list",
            "path_pattern":      r"/api/v1/pending/list",
        },
        {
            "supported_methods": ["GET"],
            "call_method":       "trigger_library_rescan",
            "path_pattern":      r"/api/v1/pending/rescan",
        },
    ]

    def initialize(self, **kwargs):
        self.name = 'pending_api'
        self.config = config.Config()

        self.params = kwargs.get("params")
        udq = UnmanicDataQueues()
        self.unmanic_data_queues = udq.get_unmanic_data_queues()

    def set_default_headers(self):
        """Set the default response header to be JSON."""
        self.set_header("Content-Type", 'application/json; charset="utf-8"')

    def get(self, path):
        self.action_route()

    def post(self, path):
        self.action_route()

    def manage_pending_tasks_list(self, *args, **kwargs):
        request_dict = json.loads(self.request.body)

        # Delete a list of tasks.
        #   (on success will continue to return the current list of tasks)
        if request_dict.get("customActionName") == "remove-from-task-list":
            if not self.delete_pending_tasks(request_dict.get("id")):
                self.write(json.dumps({"success": False}))
                return

        # Move a list of tasks to the top of the queue
        if request_dict.get("customActionName") == "move-to-top-of-task-list":
            if not pending_tasks.reorder_pending_tasks(request_dict.get("id"), "top"):
                self.write(json.dumps({"success": False}))
                return

        # Move a list of tasks to the top of the queue
        if request_dict.get("customActionName") == "move-to-bottom-of-task-list":
            if not pending_tasks.reorder_pending_tasks(request_dict.get("id"), "bottom"):
                self.write(json.dumps({"success": False}))
                return

        # Return a list of tasks based on the request JSON body
        results = pending_tasks.prepare_filtered_pending_tasks_for_table(request_dict)
        self.write(json.dumps(results))

    def trigger_library_rescan(self):
        """
        Adds a trigger ('library_scan') to the library_scanner_triggers
        data queue.
        This data queue is read by the LibraryScanner service which will
        then execute a library scan.

        :return:
        """
        # Handle request to manually trigger a rescan of the library
        # Check if we are able to start up a worker for another encoding job
        library_scanner_triggers = self.unmanic_data_queues.get('library_scanner_triggers')
        if library_scanner_triggers.full():
            self.write(json.dumps({"success": False}))
            return
        else:
            library_scanner_triggers.put('library_scan')
            self.write(json.dumps({"success": True}))
            return

    def delete_pending_tasks(self, pending_task_ids):
        """
        Deletes a list of pending tasks

        :param pending_task_ids:
        :return:
        """
        # Fetch tasks
        task_handler = task.Task()
        # Delete by ID
        return task_handler.delete_tasks_recursively(id_list=pending_task_ids)

    def reorder_pending_tasks(self, pending_task_ids, direction="top"):
        """
        Moves a list of pending tasks to either the top of the
        list of bottom depending on the provided direction.

        :param pending_task_ids:
        :param direction:
        :return:
        """
        # Fetch tasks
        task_handler = task.Task()

        return task_handler.reorder_tasks(pending_task_ids, direction)

    def add_tasks_to_pending_tasks_list(self, *args, **kwargs):
        """
        TODO: Add the given list of tasks to the pending task list

        :param args:
        :param kwargs:
        :return:
        """
        pass

    def create_task_from_path(self, *args, **kwargs):
        """
        Generate a Task object from a pathname

        :param pathname:
        :return:
        """
        request_dict = json.loads(self.request.body)

        # Fetch the abspath name
        abspath = os.path.abspath(request_dict.get("path"))

        # Ensure path exists
        if not os.path.exists(abspath):
            self.write(json.dumps({"success": False}))
            return

        self.write(json.dumps({"success": False}))
        return
        # TODO: Clean up FFmpeg code. Add this feature to v2 API without it
        # # Create a new task
        # new_task = task.Task()
        # # Run a probe on the file for current data
        # source_data = common.fetch_file_data_by_path(abspath)
        # if not new_task.create_task_by_absolute_path(abspath, self.config, source_data):
        #     # If file exists in task queue already this will return false.
        #     # Do not carry on.
        #     self.write(json.dumps({"success": False}))
        #     return
        # self.write(json.dumps({"success": True}))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.schemas.py

    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     01 Aug 2021, (11:45 AM)

    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved

           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:

           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.

           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""
from marshmallow import Schema, fields, validate


class BaseSchema(Schema):
    class Meta:
        ordered = True


# RESPONSES
# =========

class BaseSuccessSchema(BaseSchema):
    success = fields.Boolean(
        required=True,
        description='This is always "True" when a request succeeds',
        example=True,
    )


class BaseErrorSchema(BaseSchema):
    error = fields.Str(
        required=True,
        description="Return status code and reason",
    )
    messages = fields.Dict(
        required=True,
        description="Attached request body validation errors",
        example={"name": ["The thing that went wrong."]},
    )
    traceback = fields.List(
        cls_or_instance=fields.Str,
        required=False,
        description="Attached exception traceback (if developer mode is enabled)",
        example=[
            "Traceback (most recent call last):\n",
            "...",
            "json.decoder.JSONDecodeError: Expecting value: line 3 column 14 (char 45)\n"
        ],
    )


class BadRequestSchema(BaseErrorSchema):
    """STATUS_ERROR_EXTERNAL = 400"""
    error = fields.Str(
        required=True,
        description="Return status code and reason",
        example="400: Failed request schema validation",
    )


class BadEndpointSchema(BaseSchema):
    """STATUS_ERROR_ENDPOINT_NOT_FOUND = 404"""
    error = fields.Str(
        required=True,
        description="Return status code and reason",
        example="404: Endpoint not found",
    )


class BadMethodSchema(BaseSchema):
    """STATUS_ERROR_METHOD_NOT_ALLOWED = 405"""
    error = fields.Str(
        required=True,
        description="Return status code and reason",
        example="405: Method 'GET' not allowed",
    )


class InternalErrorSchema(BaseErrorSchema):
    """STATUS_ERROR_INTERNAL = 500"""
    error = fields.Str(
        required=True,
        description="Return status code and reason",
        example="500: Caught exception message",
    )


# GENERIC
# =======

class RequestTableDataSchema(BaseSchema):
    """Table request schema"""

    start = fields.Int(
        required=False,
        description="Start row number to select from",
        example=0,
        load_default=0,
    )
    length = fields.Int(
        required=False,
        description="Number of rows to select",
        example=10,
        load_default=10,
    )
    search_value = fields.Str(
        required=False,
        description="String to filter search results by",
        example="items with this text in the value",
        load_default="",
    )
    status = fields.Str(
        required=False,
        description="Filter on the status",
        example="all",
        load_default="all",
    )
    after = fields.DateTime(
        required=False,
        description="Filter entries since datetime",
        example="2022-04-07 01:45",
        allow_none=True,
    )
    before = fields.DateTime(
        required=False,
        description="Filter entries prior to datetime",
        example="2022-04-07 01:55",
        allow_none=True,
    )
    order_by = fields.Str(
        required=False,
        description="Column to order results by",
        example="finish_time",
        load_default="",
    )
    order_direction = fields.Str(
        required=False,
        description="Order direction ('asc' or 'desc')",
        example="desc",
        validate=validate.OneOf(["asc", "desc"]),
    )


class RequestTableUpdateByIdList(BaseSchema):
    """Schema for updating tables by ID"""

    id_list = fields.List(
        cls_or_instance=fields.Int,
        required=True,
        description="List of table IDs",
        example=[],
        validate=validate.Length(min=1),
    )


class TableRecordsSuccessSchema(BaseSchema):
    """Schema for table results"""

    recordsTotal = fields.Int(
        required=False,
        description="Total number of records in this table",
        example=329,
    )
    recordsFiltered = fields.Int(
        required=False,
        description="Total number of records after filters have been applied",
        example=10,
        load_default=10,
    )
    results = fields.List(
        cls_or_instance=fields.Raw,
        required=False,
        description="Results",
        example=[],
    )


class RequestDatabaseItemByIdSchema(BaseSchema):
    """Schema to request a single table item given its ID"""

    id = fields.Int(
        required=True,
        description="The ID of the table item",
        example=1,
    )


# DOCS
# ====

class DocumentContentSuccessSchema(BaseSchema):
    """Schema for updating tables by ID"""

    content = fields.List(
        cls_or_instance=fields.Str,
        required=True,
        description="Document contents read line-by-line into a list",
        example=[
            "First line\n",
            "Second line\n",
            "\n",
        ],
        validate=validate.Length(min=1),
    )


# FILEBROWSER
# ===========

class RequestDirectoryListingDataSchema(BaseSchema):
    """Schema for requesting a directory content listing"""

    current_path = fields.Str(
        example="/",
        load_default="/",
    )
    list_type = fields.Str(
        example="directories",
        load_default="all",
    )


class DirectoryListingResultsSchema(BaseSchema):
    """Schema for directory listing results returned"""

    directories = fields.List(
        cls_or_instance=fields.Dict,
        required=True,
        description="A list of directories in the given path",
        example=[
            {
                'value': "home",
                'label': "/home",
            },
            {
                'value': "tmp",
                'label': "/tmp",
            },
        ],
        validate=validate.Length(min=0),
    )
    files = fields.List(
        cls_or_instance=fields.Dict,
        required=True,
        description="A list of files in the given path",
        example=[
            {
                'value': "file1.txt",
                'label': "/file1.txt",
            },
            {
                'value': "file2.txt",
                'label': "/file2.txt",
            },
        ],
        validate=validate.Length(min=0),
    )


# HISTORY
# =======

class RequestHistoryTableDataSchema(RequestTableDataSchema):
    """Schema for requesting completed tasks from the table"""

    order_by = fields.Str(
        example="finish_time",
        load_default="finish_time",
    )


class CompletedTasksTableResultsSchema(BaseSchema):
    """Schema for completed tasks results returned by the table"""

    id = fields.Int(
        required=True,
        description="Item ID",
        example=1,
    )
    task_label = fields.Str(
        required=True,
        description="Item label",
        example="example.mp4",
    )
    task_success = fields.Boolean(
        required=True,
        description="Item success status",
        example=True,
    )
    finish_time = fields.Int(
        required=True,
        description="Item finish time",
        example=1627392616.6400812,
    )


class CompletedTasksSchema(TableRecordsSuccessSchema):
    """Schema for returning a list of completed task results"""

    successCount = fields.Int(
        required=True,
        description="Total count of times with a success status in the results list",
        example=337,
    )
    failedCount = fields.Int(
        required=True,
        description="Total count of times with a failed status in the results list",
        example=2,
    )
    results = fields.Nested(
        CompletedTasksTableResultsSchema,
        required=True,
        description="Results",
        many=True,
        validate=validate.Length(min=0),
    )


class CompletedTasksLogRequestSchema(BaseSchema):
    """Schema for requesting a task log"""

    task_id = fields.Int(
        required=True,
        description="The ID of the task",
        example=1,
    )


class CompletedTasksLogSchema(BaseSchema):
    """Schema for returning a list of completed task results"""

    command_log = fields.Str(
        required=True,
        description="Long string...",
        example='Long string...',
    )
    command_log_lines = fields.List(
        cls_or_instance=fields.Str,
        required=True,
        description="The long string broken up into an array of lines",
        example=[
            "",
            "<b>RUNNER: </b>",
            "Video Encoder H264 - libx264 [Pass #1]",
            "",
            "<b>COMMAND:</b>",
            "",
            "...",
        ],
    )


class RequestAddCompletedToPendingTasksSchema(RequestTableUpdateByIdList):
    """Schema for adding a completed task to the pending task queue"""

    library_id = fields.Int(
        required=False,
        load_default=0,
        example=1,
    )


# PENDING
# =======

class RequestPendingTableDataSchema(RequestTableDataSchema):
    """Schema for requesting pending tasks from the table"""

    order_by = fields.Str(
        example="priority",
        load_default="priority",
    )


class PendingTasksTableResultsSchema(BaseSchema):
    """Schema for pending task results returned by the table"""

    id = fields.Int(
        required=True,
        description="Item ID",
        example=1,
    )
    abspath = fields.Str(
        required=True,
        description="File absolute path",
        example="example.mp4",
    )
    priority = fields.Int(
        required=True,
        description="The current priority (higher is greater)",
        example=100,
    )
    type = fields.Str(
        required=True,
        description="The type of the pending task - local or remote",
        example="local",
    )
    status = fields.Str(
        required=True,
        description="The current status of the pending task",
        example="pending",
    )
    checksum = fields.Str(
        required=False,
        description="The uploaded file md5 checksum",
        example="5425ab3df5cdbad2e1099bb4cb963a4f",
    )
    library_id = fields.Int(
        required=False,
        description="The ID of the library for which this task was created",
        example=1,
    )
    library_name = fields.Str(
        required=False,
        description="The name of the library for which this task was created",
        example="Default",
    )


class PendingTasksSchema(TableRecordsSuccessSchema):
    """Schema for returning a list of pending task results"""

    results = fields.Nested(
        PendingTasksTableResultsSchema,
        required=True,
        description="Results",
        many=True,
        validate=validate.Length(min=0),
    )


class RequestPendingTasksReorderSchema(RequestTableUpdateByIdList):
    """Schema for moving pending items to top or bottom of table by ID"""

    position = fields.Str(
        required=True,
        description="Position to move given list of items to ('top' or 'bottom')",
        example="top",
        validate=validate.OneOf(["top", "bottom"]),
    )


class RequestPendingTaskCreateSchema(BaseSchema):
    """Schema for requesting the creation of a pending task"""

    path = fields.Str(
        required=True,
        description="The absolute path to a file",
        example="/library/TEST_FILE.mkv",
    )
    library_id = fields.Int(
        required=False,
        description="The ID of the library to append this task to",
        example=1,
    )
    library_name = fields.Str(
        required=False,
        description="The name of the library to append this task to",
        example='Default',
    )
    type = fields.Str(
        required=False,
        description="The type of pending task to create (local/remote)",
        example='local',
    )
    priority_score = fields.Int(
        required=False,
        description="Apply a priority score to the created task to either increase or decrease its position in the queue",
        example=1000,
    )


class TaskDownloadLinkSchema(BaseSchema):
    """Schema for returning a download link ID"""

    link_id = fields.Str(
        required=True,
        description="The ID used to download the file /unmanic/downloads/{link_id}",
        example="2960645c-a4e2-4b05-8866-7bd469ee9ef8",
    )


class RequestPendingTasksLibraryUpdateSchema(RequestTableUpdateByIdList):
    """Schema for updating the library for a list of created tasks"""

    library_name = fields.Str(
        required=True,
        example='Default',
    )


# PLUGINS
# =======

class RequestPluginsTableDataSchema(RequestTableDataSchema):
    """Schema for requesting plugins from the table"""

    order_by = fields.Str(
        example="name",
        load_default="name",
    )


class PluginStatusSchema(BaseSchema):
    installed = fields.Boolean(
        required=False,
        description="Is the plugin installed",
        example=True,
    )
    update_available = fields.Boolean(
        required=False,
        description="Does the plugin have an update available",
        example=True,
    )


class RequestPluginsByIdSchema(BaseSchema):
    """Schema to request data pertaining to a plugin by it's Plugin ID"""

    plugin_id = fields.Str(
        required=True,
        example="encoder_video_hevc_vaapi",
    )
    repo_id = fields.Str(
        required=False,
        description="The ID of the repository that this plugin is in",
        example="158899500680826593283708490873332175078",
    )


class PluginsMetadataResultsSchema(BaseSchema):
    """Schema for plugin metadata that will be returned by various requests """

    plugin_id = fields.Str(
        required=True,
        description="The plugin ID",
        example="encoder_video_h264_nvenc",
    )
    name = fields.Str(
        required=True,
        description="The plugin name",
        example="Video Encoder H264 - h264_nvenc",
    )
    author = fields.Str(
        required=True,
        description="The plugin author",
        example="encoder_video_h264_nvenc",
    )
    description = fields.Str(
        required=True,
        description="The plugin description",
        example="Ensure all video streams are encoded with the H264 codec using the h264_nvenc encoder.",
    )
    version = fields.Str(
        required=True,
        description="The plugin version",
        example="Josh.5",
    )
    icon = fields.Str(
        required=True,
        description="The plugin icon",
        example="https://raw.githubusercontent.com/Josh5/unmanic-plugins/master/source/encoder_video_h264_nvenc/icon.png",
    )
    tags = fields.Str(
        required=True,
        description="The plugin tags",
        example="video,encoder,ffmpeg,worker,nvenc,nvdec,nvidia",
    )
    status = fields.Nested(
        PluginStatusSchema,
        required=True,
        description="The plugin status",
    )
    changelog = fields.Str(
        required=False,
        description="The plugin changelog",
        example="[b][color=56adda]0.0.1[/color][/b]• initial version",
    )
    has_config = fields.Boolean(
        required=False,
        description="The plugin has the ability to be configured",
        example=True,
    )


class PluginsTableResultsSchema(PluginsMetadataResultsSchema):
    """Schema for pending task results returned by the table"""

    id = fields.Int(
        required=True,
        description="Item table ID",
        example=1,
    )


class PluginsDataSchema(TableRecordsSuccessSchema):
    """Schema for returning a list of plugin table results"""

    results = fields.Nested(
        PluginsTableResultsSchema,
        required=True,
        description="Results",
        many=True,
        validate=validate.Length(min=0),
    )


class RequestPluginsInfoSchema(RequestPluginsByIdSchema):
    """Schema for requesting plugins info by a given Plugin ID"""

    prefer_local = fields.Boolean(
        required=False,
        load_default=True,
        example=True,
    )
    library_id = fields.Int(
        required=False,
        load_default=0,
        example=1,
    )


class PluginsConfigInputItemSchema(BaseSchema):
    """Schema for plugin config input items"""

    key_id = fields.Str(
        required=True,
        description="The config input base64 encoded key (used for linking keys containing spaces, etc.)",
        example="c8f122656ed2acabde9b57101a4c8ec7",
    )
    key = fields.Str(
        required=True,
        description="The config input key or name",
        example="downmix_dts_hd_ma",
    )
    value = fields.Raw(
        required=True,
        description="The current value of this config input",
        example=False,
    )
    input_type = fields.Str(
        required=True,
        description="The config input type",
        example="checkbox",
    )
    label = fields.Str(
        required=True,
        description="The label used to define this config input",
        example="Downmix DTS-HD Master Audio (max 5.1 channels)?",
    )
    select_options = fields.List(
        cls_or_instance=fields.Dict,
        required=True,
        description="Additional options if the input_type is set to 'select'",
        example=[
            {
                'value': "first",
                'label': "First Option",
            },
            {
                'value': "second",
                'label': "Second Option",
            },
        ],
    )
    slider_options = fields.Dict(
        required=True,
        description="Additional options if the input_type is set to 'slider'",
        example={
            "min":    1,
            "max":    8,
            "suffix": "M"
        },
    )
    display = fields.Str(
        required=True,
        description="Should the setting input be displayed (visible, hidden)",
        example="visible",
    )


class PluginsInfoResultsSchema(PluginsMetadataResultsSchema):
    """Schema for pending task results returned by the table"""

    settings = fields.Nested(
        PluginsConfigInputItemSchema,
        required=False,
        many=True,
        description="The plugin settings",
    )


class RequestPluginsSettingsSaveSchema(BaseSchema):
    """Schema for requesting the update of a plugins settings by the plugin install ID"""

    plugin_id = fields.Str(
        required=True,
        example="encoder_video_hevc_vaapi",
    )
    settings = fields.Nested(
        PluginsConfigInputItemSchema,
        required=True,
        many=True,
        description="The plugin settings",
    )
    library_id = fields.Int(
        required=False,
        load_default=0,
        example=1,
    )


class RequestPluginsSettingsResetSchema(BaseSchema):
    """Schema for requesting the reset of a plugins settings by the plugin install ID"""

    plugin_id = fields.Str(
        required=True,
        example="encoder_video_hevc_vaapi",
    )
    library_id = fields.Int(
        required=False,
        load_default=0,
        example=1,
    )


class PluginsMetadataInstallableResultsSchema(PluginsMetadataResultsSchema):
    """Schema for plugin metadata that will be returned when fetching installable plugins """

    package_url = fields.Str(
        required=False,
        description="The plugin package download URL",
        example="https://raw.githubusercontent.com/Unmanic/unmanic-plugins/repo/plugin_id/plugin_id-1.0.0.zip",
    )
    changelog_url = fields.Str(
        required=False,
        description="The plugin package download URL",
        example="https://raw.githubusercontent.com/Unmanic/unmanic-plugins/repo/plugin_id/changelog.md",
    )
    repo_name = fields.Str(
        required=False,
        description="The name of the repository that this plugin is in",
        example="Official Repo",
    )
    repo_id = fields.Str(
        required=False,
        description="The ID of the repository that this plugin is in",
        example="158899500680826593283708490873332175078",
    )


class PluginsInstallableResultsSchema(BaseSchema):
    """Schema for installable plugins lists that are returned"""

    plugins = fields.Nested(
        PluginsMetadataInstallableResultsSchema,
        required=True,
        description="Results",
        many=True,
        validate=validate.Length(min=0),
    )


class PluginTypesResultsSchema(BaseSchema):
    """Schema for installable plugins lists that are returned"""

    results = fields.List(
        cls_or_instance=fields.Str,
        required=True,
        description="List of Plugin Type IDs supported by this installation",
        example=[
            "library_management.file_test",
            "postprocessor.file_move",
            "postprocessor.task_result",
            "worker.process_item"
        ],
    )


class RequestPluginsFlowByPluginTypeSchema(BaseSchema):
    """Schema to request the plugin flow of a given plugin type"""

    plugin_type = fields.Str(
        required=True,
        example="library_management.file_test",
    )
    library_id = fields.Int(
        required=False,
        load_default=1,
        example=1,
    )


class PluginFlowDataResultsSchema(BaseSchema):
    """Schema for plugin flow data items"""

    plugin_id = fields.Str(
        required=True,
        description="The plugin ID",
        example="encoder_video_h264_nvenc",
    )
    name = fields.Str(
        required=True,
        description="The plugin name",
        example="Video Encoder H264 - h264_nvenc",
    )
    author = fields.Str(
        required=True,
        description="The plugin author",
        example="encoder_video_h264_nvenc",
    )
    description = fields.Str(
        required=True,
        description="The plugin description",
        example="Ensure all video streams are encoded with the H264 codec using the h264_nvenc encoder.",
    )
    version = fields.Str(
        required=True,
        description="The plugin version",
        example="Josh.5",
    )
    icon = fields.Str(
        required=True,
        description="The plugin icon",
        example="https://raw.githubusercontent.com/Josh5/unmanic-plugins/master/source/encoder_video_h264_nvenc/icon.png",
    )


class PluginFlowResultsSchema(BaseSchema):
    """Schema for returned plugin flow list"""

    results = fields.Nested(
        PluginFlowDataResultsSchema,
        required=True,
        description="Results",
        many=True,
        validate=validate.Length(min=0),
    )


class RequestSavingPluginsFlowByPluginTypeSchema(RequestPluginsFlowByPluginTypeSchema):
    """Schema to request saving the plugin flow of a given plugin type"""

    plugin_flow = fields.Nested(
        PluginFlowDataResultsSchema,
        required=True,
        description="Saved flow",
        many=True,
        validate=validate.Length(min=1),
    )
    library_id = fields.Int(
        required=False,
        load_default=1,
        example=1,
    )


class PluginReposMetadataResultsSchema(BaseSchema):
    """Schema for plugin repo metadata that will be returned when fetching repo lists"""

    id = fields.Str(
        required=True,
        description="The plugin repo ID",
        example="repository.josh5",
    )
    name = fields.Str(
        required=True,
        description="The plugin repo name",
        example="Josh.5 Development Plugins for Unmanic",
    )
    icon = fields.Str(
        required=True,
        description="The plugin repo icon",
        example="https://raw.githubusercontent.com/Josh5/unmanic-plugins/master/icon.png",
    )
    path = fields.Str(
        required=True,
        description="The plugin repo URL path",
        example="https://raw.githubusercontent.com/Josh5/unmanic-plugins/repo/repo.json",
    )


class RequestUpdatePluginReposListSchema(BaseSchema):
    """Schema to request an update of the plugin repos list"""

    repos_list = fields.List(
        cls_or_instance=fields.Str,
        required=True,
        description="A list of repost to save",
        example=[
            'https://raw.githubusercontent.com/Josh5/unmanic-plugins/repo/repo.json',
        ],
        validate=validate.Length(min=0),
    )


class PluginReposListResultsSchema(BaseSchema):
    """Schema for plugin repo lists that are returned"""

    repos = fields.Nested(
        PluginReposMetadataResultsSchema,
        required=True,
        description="Results",
        many=True,
        validate=validate.Length(min=0),
    )


class PluginsDataPanelTypesDataSchema(BaseSchema):
    """Schema for returning a list of data panel plugins results"""

    results = fields.Nested(
        PluginFlowDataResultsSchema,
        required=True,
        description="Results",
        many=True,
        validate=validate.Length(min=0),
    )


# SESSION
# =======

class SessionStateSuccessSchema(BaseSchema):
    """Schema for returning session data"""

    level = fields.Int(
        required=True,
        description="User level",
        example=0,
    )
    picture_uri = fields.Str(
        required=False,
        description="User picture",
        example="https://c8.patreon.com/2/200/561356054",
    )
    name = fields.Str(
        required=False,
        description="User name",
        example="ExampleUsername123",
    )
    email = fields.Str(
        required=False,
        description="User email",
        example="example@gmail.com",
    )
    created = fields.Number(
        required=False,
        description="Session time created",
        example=1627793093.676484,
    )
    uuid = fields.Str(
        required=True,
        description="Installation uuid",
        example="b429fcc7-9ce1-bcb3-2b8a-b094747f226e",
    )


# SETTINGS
# ========

class SettingsReadAndWriteSchema(BaseSchema):
    """Schema to request the current settings"""

    settings = fields.Dict(
        required=True,
        description="The current settings",
        example={
            "ui_port":                    8888,
            "debugging":                  False,
            "library_path":               "/library",
            "enable_library_scanner":     False,
            "schedule_full_scan_minutes": 1440,
            "follow_symlinks":            True,
            "run_full_scan_on_start":     False,
            "cache_path":                 "/tmp/unmanic"
        },
    )


class SettingsSystemConfigSchema(BaseSchema):
    """Schema to display the current system configuration"""

    configuration = fields.Dict(
        required=True,
        description="The current system configuration",
        example={},
    )


class WorkerEventScheduleResultsSchema(BaseSchema):
    """Schema for worker status results"""

    repetition = fields.Str(
        required=True,
        description="",
        example="daily",
    )
    schedule_task = fields.Str(
        required=True,
        description="The type of task. ['count', 'pause', 'resume']",
        example="count",
    )
    schedule_time = fields.Str(
        required=True,
        description="",
        example="The time when the task should be executed on",
    )
    schedule_worker_count = fields.Int(
        required=False,
        description="The worker count to set (only valid if schedule_task is count)",
        example=4,
    )


class SettingsWorkerGroupConfigSchema(BaseSchema):
    """Schema to display the config of a single worker group"""

    id = fields.Int(
        required=True,
        description="",
        example=1,
        allow_none=True,
    )
    locked = fields.Boolean(
        required=True,
        description="If the worker group is locked and cannot be deleted",
        example=False,
    )
    name = fields.Str(
        required=True,
        description="The name of the worker group",
        example="Default Group",
    )
    number_of_workers = fields.Int(
        required=True,
        description="The number of workers in this group",
        example=3,
    )
    worker_event_schedules = fields.Nested(
        WorkerEventScheduleResultsSchema,
        required=True,
        description="Any scheduled evenets for this worker group",
        many=True,
        validate=validate.Length(min=0),
    )
    tags = fields.List(
        cls_or_instance=fields.Str,
        required=True,
        description="A list of tags associated with this worker",
        example=['GPU', 'priority'],
    )


class WorkerGroupsListSchema(BaseSchema):
    """Schema to list all worker groups"""

    worker_groups = fields.Nested(
        SettingsWorkerGroupConfigSchema,
        required=True,
        description="Results",
        many=True,
        validate=validate.Length(min=0),
    )


class RequestSettingsRemoteInstallationAddressValidationSchema(BaseSchema):
    """Schema to request validation of remote installation address"""

    address = fields.Str(
        required=True,
        description="The address of the remote installation",
        example="192.168.1.2:8888",
    )
    auth = fields.Str(
        required=False,
        description="Authentication type",
        example="Basic",
        allow_none=True,
    )
    username = fields.Str(
        required=False,
        description="An optional username",
        example="foo",
        allow_none=True,
    )
    password = fields.Str(
        required=False,
        description="An optional password",
        example="bar",
        allow_none=True,
    )


class SettingsRemoteInstallationDataSchema(BaseSchema):
    """Schema to display the data from the remote installation"""

    installation = fields.Dict(
        required=True,
        description="The data from the remote installation",
        example={},
    )


class RequestRemoteInstallationLinkConfigSchema(BaseSchema):
    """Schema to request a single remote installation link configuration given its UUID"""

    uuid = fields.Str(
        required=True,
        description="The uuid of the remote installation",
        example="7cd35429-76ab-4a29-8649-8c91236b5f8b",
    )


class SettingsRemoteInstallationLinkConfigSchema(BaseSchema):
    """Schema to display the data from the remote installation"""

    link_config = fields.Dict(
        required=True,
        description="The configuration for the remote installation link",
        example={
            "address":                         "10.0.0.2:8888",
            "auth":                            "None",
            "username":                        "",
            "password":                        "",
            "available":                       True,
            "name":                            "API schema generated",
            "version":                         "0.1.3",
            "last_updated":                    1636166593.013826,
            "enable_receiving_tasks":          False,
            "enable_sending_tasks":            False,
            "enable_task_preloading":          True,
            "enable_distributed_worker_count": False,
            "preloading_count":                2,
            "enable_checksum_validation":      False,
            "enable_config_missing_libraries": False,
        },
    )
    distributed_worker_count_target = fields.Int(
        required=False,
        description="The target count of workers to be distributed across any configured linked installations",
        example=4,
    )


class LibraryResultsSchema(BaseSchema):
    """Schema for library results"""

    id = fields.Int(
        required=True,
        description="",
        example=1,
    )
    name = fields.Str(
        required=True,
        description="The name of the library",
        example="Default",
    )
    path = fields.Str(
        required=True,
        description="The library path",
        example="/library",
    )
    locked = fields.Boolean(
        required=True,
        description="If the library is locked and cannot be deleted",
        example=False,
    )
    enable_remote_only = fields.Boolean(
        required=True,
        description="If the library is configured for remote files only",
        example=False,
    )
    enable_scanner = fields.Boolean(
        required=True,
        description="If the library is configured to execute library scans",
        example=False,
    )
    enable_inotify = fields.Boolean(
        required=True,
        description="If the library is configured to monitor for file changes",
        example=False,
    )
    tags = fields.List(
        cls_or_instance=fields.Str,
        required=True,
        description="A list of tags associated with this library",
        example=['GPU', 'priority'],
    )


class SettingsLibrariesListSchema(BaseSchema):
    """Schema to list all libraries"""

    libraries = fields.Nested(
        LibraryResultsSchema,
        required=True,
        description="Results",
        many=True,
        validate=validate.Length(min=1),
    )


class RequestLibraryByIdSchema(BaseSchema):
    """Schema to request a single library given its ID"""

    id = fields.Int(
        required=True,
        description="The ID of the library",
        example=1,
    )


class SettingsLibraryConfigReadAndWriteSchema(BaseSchema):
    """Schema to display the data from the remote installation"""

    library_config = fields.Dict(
        required=True,
        description="The library configuration",
        example={
            "id":             1,
            "name":           "Default",
            "path":           "/library",
            "enable_scanner": False,
            "enable_inotify": False,
            "priority_score": 0,
            "tags":           [],
        },
    )

    plugins = fields.Dict(
        required=False,
        description="The library's enabled plugins",
        example={
            "enabled_plugins": [
                {
                    "library_id":  1,
                    "plugin_id":   "notify_plex",
                    "name":        "Notify Plex",
                    "description": "Notify Plex on completion of a task.",
                    "icon":        "https://raw.githubusercontent.com/Josh5/unmanic.plugin.notify_plex/master/icon.png"
                }
            ]
        },
    )


class SettingsLibraryPluginConfigExportSchema(BaseSchema):
    """Schema for exporting a library's plugin config"""

    plugins = fields.Dict(
        required=True,
        description="The library's enabled plugins",
        example={
            "enabled_plugins": [
                {
                    "library_id":  1,
                    "plugin_id":   "encoder_audio_ac3",
                    "name":        "Audio Encoder AC3",
                    "description": "Ensure all audio streams are encoded with the AC3 codec using the native FFmpeg ac3 encoder.",
                    "icon":        "https://raw.githubusercontent.com/Josh5/unmanic.plugin.encoder_audio_ac3/master/icon.png"
                }
            ],
            "plugin_flow":     {
                "library_management.file_test": [
                    {
                        "plugin_id":   "encoder_audio_ac3",
                        "name":        "Audio Encoder AC3",
                        "author":      "Josh.5",
                        "description": "Ensure all audio streams are encoded with the AC3 codec using the native FFmpeg ac3 encoder.",
                        "version":     "0.0.2",
                        "icon":        "https://raw.githubusercontent.com/Josh5/unmanic.plugin.encoder_audio_ac3/master/icon.png"
                    }
                ],
                "worker.process_item":          [
                    {
                        "plugin_id":   "encoder_audio_ac3",
                        "name":        "Audio Encoder AC3",
                        "author":      "Josh.5",
                        "description": "Ensure all audio streams are encoded with the AC3 codec using the native FFmpeg ac3 encoder.",
                        "version":     "0.0.2",
                        "icon":        "https://raw.githubusercontent.com/Josh5/unmanic.plugin.encoder_audio_ac3/master/icon.png"
                    }
                ],
                "postprocessor.file_move":      [],
                "postprocessor.task_result":    []
            }
        },
    )

    library_config = fields.Dict(
        required=False,
        description="The library configuration",
        example={
            "id":             1,
            "name":           "Default",
            "path":           "/library",
            "enable_scanner": False,
            "enable_inotify": False,
            "priority_score": 0,
            "tags":           [],
        },
    )


class SettingsLibraryPluginConfigImportSchema(SettingsLibraryPluginConfigExportSchema):
    """Schema for import a library's plugin config"""

    library_id = fields.Int(
        required=True,
        example=1,
    )


# VERSION
# =======

class VersionReadSuccessSchema(BaseSchema):
    """Schema for returning the application version"""

    version = fields.Str(
        required=True,
        description="Application version",
        example="1.0.0",
    )


# WORKERS
# =======

class RequestWorkerByIdSchema(BaseSchema):
    """Schema to request a worker by the worker's ID"""

    worker_id = fields.Str(
        required=True,
        example="1",
    )


class WorkerStatusResultsSchema(BaseSchema):
    """Schema for worker status results"""

    id = fields.Str(
        required=True,
        description="",
        example="W0",
    )
    name = fields.Str(
        required=True,
        description="",
        example="Worker-W0",
    )
    idle = fields.Boolean(
        required=True,
        description="Flag - is worker idle",
        example=True,
    )
    paused = fields.Boolean(
        required=True,
        description="Flag - is worker paused",
        example=False,
    )
    start_time = fields.Str(
        required=True,
        description="The time when this worker started processing a task",
        example="1635746377.0021548",
        allow_none=True,
    )
    current_file = fields.Str(
        required=True,
        description="The basename of the file currently being processed",
        example="file.mp4",
    )
    current_task = fields.Int(
        required=True,
        description="The Task ID",
        example=1,
        allow_none=True,
    )
    worker_log_tail = fields.List(
        cls_or_instance=fields.Str,
        required=True,
        description="The log lines produced by the worker",
        example=[
            "\n\nRUNNER: \nRemux Video Files [Pass #1]\n\n",
            "\nExecuting plugin runner... Please wait",
            "\nRunner did not request to execute a command",
            "\n\nNo Plugin requested to run commands for this file '/tmp/unmanic/unmanic_remote_pending_library-1635746225.3336523/file.mp4'"
        ],
        validate=validate.Length(min=0),
    )
    runners_info = fields.Dict(
        required=True,
        description="The status of the plugin runner currently processing the file",
        example={
            "video_remuxer": {
                "plugin_id":   "video_remuxer",
                "status":      "complete",
                "name":        "Remux Video Files",
                "author":      "Josh.5",
                "version":     "0.0.5",
                "icon":        "https://raw.githubusercontent.com/Josh5/unmanic.plugin.video_remuxer/master/icon.png",
                "description": "Remux a video file to the configured container",
                "success":     True
            }
        },
    )
    subprocess = fields.Dict(
        required=True,
        description="The status of the process currently being executed",
        example={
            "pid":     140408939493120,
            "percent": "None",
            "elapsed": "None"
        },
    )


class WorkerStatusSuccessSchema(BaseSchema):
    """Schema for returning the status of all workers"""

    workers_status = fields.Nested(
        WorkerStatusResultsSchema,
        required=True,
        description="Results",
        many=True,
        validate=validate.Length(min=0),
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.plugins_api.py

    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     03 Mar 2021, (12:10 PM)

    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved

           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:

           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.

           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""
import hashlib
import json
import tornado.log

from unmanic.libs.plugins import PluginsHandler
from unmanic.libs.uiserver import UnmanicDataQueues
from unmanic.libs.unplugins import PluginExecutor
from unmanic.webserver.api_v1.base_api_handler import BaseApiHandler


class ApiPluginsHandler(BaseApiHandler):
    name = None
    params = None
    unmanic_data_queues = None

    routes = [
        {
            "supported_methods": ["GET"],
            "call_method":       "get_plugin_list",
            "path_pattern":      r"/api/v1/plugins/list",
        },
        {
            "supported_methods": ["POST"],
            "call_method":       "install_plugin_by_id",
            "path_pattern":      r"/api/v1/plugins/install",
        },
        {
            "supported_methods": ["GET"],
            "call_method":       "get_repo_list",
            "path_pattern":      r"/api/v1/plugins/repos/list",
        },
        {
            "supported_methods": ["POST"],
            "call_method":       "update_repo_list",
            "path_pattern":      r"/api/v1/plugins/repos/update",
        },
        {
            "supported_methods": ["GET"],
            "call_method":       "update_repos",
            "path_pattern":      r"/api/v1/plugins/repos/fetch",
        },
    ]

    def initialize(self, **kwargs):
        self.name = 'plugins_api'
        self.params = kwargs.get("params")
        udq = UnmanicDataQueues()
        self.unmanic_data_queues = udq.get_unmanic_data_queues()

    def set_default_headers(self):
        """Set the default response header to be JSON."""
        self.set_header("Content-Type", 'application/json; charset="utf-8"')

    def get(self, path):
        self.action_route()

    def post(self, path):
        self.action_route()

    def prepare_filtered_plugins(self, request_dict):
        """
        Returns a object of records filtered and sorted
        according to the provided request.

        :param request_dict:
        :return:
        """

        # Generate filters for query
        start = request_dict.get('start')
        length = request_dict.get('length')

        search = request_dict.get('search')
        search_value = search.get("value")

        # Force sort order always by ID desc
        order = [
            {
                "column": 'name',
                "dir":    'asc',
            }
        ]

        # Fetch Plugins
        plugins = PluginsHandler()
        # Get total count
        records_total_count = plugins.get_total_plugin_list_count()
        # Get quantity after filters (without pagination)
        records_filtered_count = plugins.get_plugin_list_filtered_and_sorted(order=order, start=0, length=0,
                                                                             search_value=search_value).count()
        # Get filtered/sorted results
        plugin_results = plugins.get_plugin_list_filtered_and_sorted(order=order, start=start, length=length,
                                                                     search_value=search_value)

        # Build return data
        return_data = {
            "draw":            request_dict.get('draw'),
            "recordsTotal":    records_total_count,
            "recordsFiltered": records_filtered_count,
            "successCount":    0,
            "failedCount":     0,
            "data":            []
        }

        # Iterate over plugins and append them to the plugin data
        for plugin_result in plugin_results:
            # Set plugin status
            plugin_status = {
                "enabled":          plugin_result.get('enabled'),
                "update_available": plugin_result.get('update_available'),
            }
            # Set params as required in template
            item = {
                'id':          plugin_result.get('id'),
                'plugin_id':   plugin_result.get('plugin_id'),
                'icon':        plugin_result.get('icon'),
                'name':        plugin_result.get('name'),
                'description': plugin_result.get('description'),
                'tags':        plugin_result.get('tags'),
                'author':      plugin_result.get('author'),
                'version':     plugin_result.get('version'),
                'status':      plugin_status,
                'selected':    False,
            }
            return_data["data"].append(item)

        # Return results
        return return_data

    def update_repo_list(self, *args, **kwargs):
        repos_list = self.get_argument('repos_list')

        if repos_list:
            repos_list = repos_list.splitlines()

        try:
            plugins = PluginsHandler()
            plugins.set_plugin_repos(repos_list)
        except Exception as e:
            tornado.log.app_log.exception("Exception in updating repo list - {}".format(str(e)), exc_info=True)

        self.get_repo_list()

    def get_repo_list(self):
        try:
            plugins = PluginsHandler()
            # Fetch the data again from the database
            current_repos = plugins.get_plugin_repos()

            # Remove the default plugin repo from the list
            return_repos = []
            default_repo = plugins.get_default_repo()
            for repo in current_repos:
                if not repo.get("path").startswith(default_repo):
                    return_repos.append(repo)

            # Return success
            self.write(json.dumps({"success": True, "repos": return_repos}))
        except Exception as e:
            tornado.log.app_log.exception("Exception in fetching the current repo list - {}".format(str(e)), exc_info=True)

            # Return failure
            self.write(json.dumps({"success": False}))

    def update_repos(self, *args, **kwargs):
        plugins = PluginsHandler()
        if plugins.update_plugin_repos():
            # Return success
            self.write(json.dumps({"success": True}))
        else:
            # Return failure
            self.write(json.dumps({"success": False}))

    def get_plugin_list(self, *args, **kwargs):
        plugins = PluginsHandler()
        # Fetch a list of plugin data cached locally
        plugin_list = plugins.get_installable_plugins_list()
        self.write(json.dumps({"success": True, "plugins": plugin_list}))

    def install_plugin_by_id(self, *args, **kwargs):
        plugin_id = self.get_argument('plugin_id')

        # Fetch a list of plugin data cached locally
        plugins = PluginsHandler()
        success = plugins.install_plugin_by_id(plugin_id)

        if success:
            # Return success
            self.write(json.dumps({"success": True}))
        else:
            # Return failure
            self.write(json.dumps({"success": False}))

    def __get_plugin_changelog(self, plugin_id):
        """
        Given a plugin ID , return a list of lines read from the plugin's changelog

        :param plugin_id:
        :return:
        """
        # Fetch plugin changelog
        plugin_executor = PluginExecutor()
        return plugin_executor.get_plugin_changelog(plugin_id)

    def __get_plugin_long_description(self, plugin_id):
        """
        Given a plugin ID , return a list of lines read from the plugin's changelog

        :param plugin_id:
        :return:
        """
        # Fetch plugin changelog
        plugin_executor = PluginExecutor()
        return plugin_executor.get_plugin_long_description(plugin_id)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.plugin_type_base.py

    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     05 Mar 2021, (8:09 PM)

    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved

           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:

           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.

           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""
import inspect


class PluginType(object):
    """
    PluginType

    Generic configuration and methods used across all plugin types
    """
    name = ''
    runner = ''
    runner_docstring = ''
    data_schema = {}
    test_data = {}

    def plugin_type_name(self):
        """
        Return the plugin runner string

        :return:
        """
        return self.name

    def plugin_runner(self):
        """
        Return the plugin runner string

        :return:
        """
        return self.runner

    def plugin_runner_docstring(self):
        """
        Return the plugin runner docstring

        :return:
        """
        return self.runner_docstring

    def get_plugin_runner_function(self, plugin_module):
        plugin_runner = self.plugin_runner()
        # Check if this module contains the given plugin type runner function
        if hasattr(plugin_module, plugin_runner):
            # If it does, add it to the plugin_modules list
            return getattr(plugin_module, plugin_runner)
        return None

    def get_data_schema(self):
        """
        Return the plugin data schema dictionary

        :return:
        """
        return self.data_schema

    def get_test_data(self):
        """
        Return the plugin test data dictionary

        :return:
        """
        return self.test_data

    def __data_schema_test_data(self, plugin_id, plugin_runner, result_data, data_schema, data_tree="/"):
        """
        Ensure the test data returned is valid according to the schema

        :param plugin_id:
        :param plugin_runner:
        :param result_data:
        :param data_schema:
        :param data_tree:
        :return:
        """

        def test_data_type(provided_data, expected_data_type):
            # Callable functions are best tested with the callable function
            # Everything else should be tested with the isinstance function
            if expected_data_type == 'callable':
                if callable(provided_data):
                    return True
            elif isinstance(provided_data, expected_data_type):
                return True
            return False

        errors = []
        if not isinstance(result_data, dict):
            # This runner function is not returning anything
            error = "Plugin '{0} - {1}()' has failed to return any output data.".format(plugin_id, plugin_runner, data_tree)
            errors.append(error)
            return errors
        for key in data_schema:
            schema_meta = data_schema.get(key)
            if schema_meta.get('required'):
                # Ensure the required item is present in result_data
                if not key in result_data:
                    error = "Plugin '{0} - {1}()' is missing required key '{2}{3}' in the output data.".format(plugin_id,
                                                                                                               plugin_runner,
                                                                                                               data_tree, key)
                    errors.append(error)

            # Ensure that data present is of the correct type
            # Recursively check for children elements
            data_type = schema_meta.get('type')
            if key in result_data:
                child_data = result_data.get(key)

                # Test that the data is of the correct type
                # Types can be multiple things for some plugin runners. If type is a list of types,
                #   iterate over that list and test all types.
                correct_type = False
                if isinstance(data_type, list):
                    for dt in data_type:
                        if test_data_type(child_data, dt):
                            correct_type = True
                            break
                else:
                    correct_type = test_data_type(child_data, data_type)

                # If data is not of the correct type, then append the error message
                if not correct_type:
                    error = "Plugin '{0} - {1}()' output data returned incorrect data type in key '{2}{3}'. " \
                            "Expected '{4}', but received '{5}'.".format(plugin_id, plugin_runner,
                                                                         data_tree, key, data_type,
                                                                         type(result_data.get(key)))
                    errors.append(error)
                # Check if data_schema has children
                children_data_schema = schema_meta.get('children')
                if children_data_schema:
                    child_data_tree = "{}{}>".format(data_tree, key)
                    errors += self.__data_schema_test_data(plugin_id, plugin_runner, child_data, children_data_schema,
                                                           data_tree=child_data_tree)

        return errors

    def run_data_schema_tests(self, plugin_id, plugin_module, test_data=None):
        """
        With a given set of test data, this method tests the provided
        plugin module's data output against the schema dictionary.

        :param plugin_module:
        :return:
        """
        plugin_runner = self.plugin_runner()
        plugin_runner_function = self.get_plugin_runner_function(plugin_module)

        # Get test data
        if not test_data:
            test_data = self.get_test_data()
        test_data_copy = test_data.copy()

        # Get data schema
        data_schema = self.get_data_schema()

        # Execute plugin function
        run_count = 0
        while run_count < 2:
            plugin_runner_function(test_data_copy)
            if not test_data_copy.get('repeat', False):
                break
            run_count += 1

        # Ensure the modified test data is valid according to the schema
        errors = self.__data_schema_test_data(plugin_id, plugin_runner, test_data_copy, data_schema)

        return errors



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.session_api.py

    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     10 Mar 2021, (7:14 PM)

    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved

           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:

           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.

           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""
import json
import tornado.log

from unmanic.libs import session
from unmanic.libs.uiserver import UnmanicDataQueues
from unmanic.webserver.api_v1.base_api_handler import BaseApiHandler


class ApiSessionHandler(BaseApiHandler):
    name = None
    session = None
    config = None
    params = None
    unmanic_data_queues = None

    routes = [
        {
            "supported_methods": ["GET"],
            "call_method":       "get_sign_out_url",
            "path_pattern":      r"/api/v1/session/unmanic-sign-out-url",
        },
        {
            "supported_methods": ["GET"],
            "call_method":       "get_patreon_login_url",
            "path_pattern":      r"/api/v1/session/unmanic-patreon-login-url",
        },
        {
            "supported_methods": ["GET"],
            "call_method":       "get_github_login_url",
            "path_pattern":      r"/api/v1/session/unmanic-github-login-url",
        },
        {
            "supported_methods": ["GET"],
            "call_method":       "get_discord_login_url",
            "path_pattern":      r"/api/v1/session/unmanic-discord-login-url",
        },
        {
            "supported_methods": ["GET"],
            "call_method":       "get_patreon_page",
            "path_pattern":      r"/api/v1/session/unmanic-patreon-page",
        },
    ]

    def initialize(self, **kwargs):
        self.name = 'plugins_api'
        self.session = session.Session()
        self.params = kwargs.get("params")
        udq = UnmanicDataQueues()
        self.unmanic_data_queues = udq.get_unmanic_data_queues()

    def set_default_headers(self):
        """Set the default response header to be JSON."""
        self.set_header("Content-Type", 'application/json; charset="utf-8"')

    def get(self, path):
        self.action_route()

    def post(self, path):
        self.action_route()

    def get_sign_out_url(self, *args, **kwargs):
        uuid = self.session.get_installation_uuid()
        sign_out_url = self.session.get_sign_out_url()
        if not sign_out_url:
            self.write(json.dumps({"success": False}))
            return
        else:
            self.write(json.dumps({
                "success": True,
                "uuid":    uuid,
                "data":    {
                    "url": sign_out_url,
                }
            }))
            return

    def get_patreon_login_url(self, *args, **kwargs):
        uuid = self.session.get_installation_uuid()
        patreon_oauth_url = self.session.get_patreon_login_url()
        if not patreon_oauth_url:
            self.write(json.dumps({"success": False}))
            return
        else:
            self.write(json.dumps({
                "success": True,
                "uuid":    uuid,
                "data":    {
                    "url": patreon_oauth_url,
                }
            }))
            return

    def get_github_login_url(self, *args, **kwargs):
        uuid = self.session.get_installation_uuid()
        github_oauth_url = self.session.get_github_login_url()
        if not github_oauth_url:
            self.write(json.dumps({"success": False}))
            return
        else:
            self.write(json.dumps({
                "success": True,
                "uuid":    uuid,
                "data":    {
                    "url": github_oauth_url,
                }
            }))
            return

    def get_discord_login_url(self, *args, **kwargs):
        uuid = self.session.get_installation_uuid()
        discord_oauth_url = self.session.get_discord_login_url()
        if not discord_oauth_url:
            self.write(json.dumps({"success": False}))
            return
        else:
            self.write(json.dumps({
                "success": True,
                "uuid":    uuid,
                "data":    {
                    "url": discord_oauth_url,
                }
            }))
            return

    def get_patreon_page(self, *args, **kwargs):
        uuid = self.session.get_installation_uuid()
        patreon_sponsor_page_data = self.session.get_patreon_sponsor_page()
        if not patreon_sponsor_page_data:
            self.write(json.dumps({"success": False}))
            return
        else:
            sponsor_page = patreon_sponsor_page_data.get("sponsor_page")
            self.write(json.dumps({
                "success": True,
                "uuid":    uuid,
                "data":    {
                    "sponsor_page": sponsor_page,
                }
            }))
            return



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.basemodel.py
 
    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     22 Jun 2019, (1:58 PM)
 
    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved
 
           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:
  
           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.
  
           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""

from peewee import *
from playhouse.sqliteq import SqliteQueueDatabase
from datetime import datetime
from base64 import b64decode
from playhouse.shortcuts import model_to_dict, dict_to_model

# Do not initialise the database until the model is called.
from unmanic.libs.singleton import SingletonType

db = DatabaseProxy()  # Create a proxy for our db.

# Stipulate date and time formats
DATE_FORMAT = '%Y-%m-%d'
TIME_FORMAT = '%H:%M:%S'
TIME_FORMAT_ALT = '{}.%f'.format(TIME_FORMAT)
DATETIME_BASE = '{}T{}'
DATETIME_FORMAT = DATETIME_BASE.format(DATE_FORMAT, TIME_FORMAT)
DATETIME_FORMAT_ALT = DATETIME_BASE.format(DATE_FORMAT, TIME_FORMAT_ALT)


def strpdatetime(string):
    """
    Parses the datetime from a string.

    """
    if string is not None:
        try:
            return datetime.strptime(string, DATETIME_FORMAT)
        except ValueError:
            return datetime.strptime(string, DATETIME_FORMAT_ALT)


def strpdate(string):
    """
    Parses the date from a string.

    """
    if string is not None:
        return datetime.strptime(string, DATE_FORMAT).date()


def strptime(string):
    """
    Parses the time from a string.

    """
    if string is not None:
        try:
            return datetime.strptime(string, TIME_FORMAT).time()
        except ValueError:
            return datetime.strptime(string, TIME_FORMAT_ALT).time()


class NoSuchFieldError(TypeError):
    """
    Indicates that the field does not exist in this model

    """
    pass


class NullError(TypeError):
    """
    Indicates that the respective field was set to NULL but must not be NULL.

    """
    pass


class Database:
    """
    Database

    Select a database to connect to
    """

    @staticmethod
    def select_database(config):
        # Based on configuration, use a different database.
        if config['TYPE'] == 'SQLITE':
            # use SqliteQueueDatabase
            database = SqliteQueueDatabase(
                config['FILE'],
                use_gevent=False,
                autostart=True,
                queue_max_size=None,
                results_timeout=15.0,
                pragmas=(
                    ('foreign_keys', 1),
                    ('journal_mode', 'wal'),
                )
            )

            db.initialize(database)
            db.connect()
        return db


class BaseModel(Model):
    """
    BaseModel

    Generic configuration and methods used across all Model classes
    """

    class Meta:
        database = db

    def get_fields(self):
        """
        Return a dictionary of this models field metadata

        :return:
        """
        return self._meta.fields

    def get_current_field_values_dict(self):
        """
        Return a dictionary of this models fields and their current values

        :return:
        """
        return self.__data__

    def parse_field_value_by_type(self, field_id, value):
        """
        Fetches the field type for this field.
        Return the passed value with the correct type.

        :param field_id:
        :param value:
        :return:
        """
        model_fields = self.get_fields()

        if field_id not in model_fields.keys():
            raise NoSuchFieldError()

        # Set the field from model
        field = model_fields[field_id]

        if value is None:
            if not field.null:
                raise NullError()
            return value

        if isinstance(field, BooleanField):
            if isinstance(value, (bool, int)):
                return bool(value)
            elif value.lower() in ['t', 'true', '1']:
                return True
            elif value.lower() in ['f', 'false', '0']:
                return False
            return False
        elif isinstance(field, IntegerField):
            return int(value)
        elif isinstance(field, FloatField):
            return float(value)
        elif isinstance(field, DecimalField):
            return float(value)
        elif isinstance(field, DateTimeField):
            return strpdatetime(value)
        elif isinstance(field, DateField):
            return strpdate(value)
        elif isinstance(field, TimeField):
            return strptime(value)
        elif isinstance(field, BlobField):
            return b64decode(value)

        return value

    def model_to_dict(self):
        """
        Retrieve all related objects recursively and
        then converts the resulting objects to a dictionary.

        :return:
        """
        return model_to_dict(self, backrefs=True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.uiserver.py

    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     06 Dec 2018, (7:21 AM)

    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved

           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:

           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.

           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""

import os
import socket
import threading
import asyncio
import logging
from queue import Queue

from tornado.httpserver import HTTPServer
from tornado.ioloop import IOLoop
from tornado.platform.asyncio import AnyThreadEventLoopPolicy
from tornado.routing import PathMatches
from tornado.template import Loader
from tornado.web import Application, StaticFileHandler, RedirectHandler

from unmanic import config
from unmanic.libs import common
from unmanic.libs.singleton import SingletonType
from unmanic.webserver.downloads import DownloadsHandler

public_directory = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "webserver", "public"))
tornado_settings = {
    'template_loader': Loader(public_directory),
    'static_css':      os.path.join(public_directory, "css"),
    'static_fonts':    os.path.join(public_directory, "fonts"),
    'static_icons':    os.path.join(public_directory, "icons"),
    'static_img':      os.path.join(public_directory, "img"),
    'static_js':       os.path.join(public_directory, "js"),
    'debug':           True,
    'autoreload':      False,
}


class FrontendPushMessages(Queue, metaclass=SingletonType):
    """
    Handles messages passed to the frontend.

    Messages are sent as objects. These objects require the following fields:
        - 'id'          : A unique ID of the message. Prevents messages duplication
        - 'type'        : The type of message - 'error', 'warning', 'success', or 'info'
        - 'code'        : A code to represent an I18n string for the frontend to display
        - 'message'     : Additional message string that can be appended to the I18n string displayed on the frontend.
        - 'timeout'     : The timeout for this message. If set to 0, then the message will persist until manually dismissed.

    """

    def _init(self, maxsize):
        self.all_items = set()
        Queue._init(self, maxsize)

    def put(self, item):
        # Ensure received item is valid
        self.__validate_item(item)
        # If it is not already in message list, add it to the list and the queue
        if item.get('id') not in self.all_items:
            self.all_items.add(item.get('id'))
            self.add_to_queue(item)

    def add_to_queue(self, item, block=True, timeout=None):
        Queue.put(self, item, block, timeout)

    @staticmethod
    def __validate_item(item):
        # Ensure all required keys are present
        for key in ['id', 'type', 'code', 'message', 'timeout']:
            if key not in item:
                raise Exception("Frontend message item incorrectly formatted. Missing key: '{}'".format(key))

        # Ensure the given type is valid
        if item.get('type') not in ['error', 'warning', 'success', 'info', 'status']:
            raise Exception(
                "Frontend message item's code must be in ['error', 'warning', 'success', 'info', 'status']. Received '{}'".format(
                    item.get('type')
                )
            )
        return True

    def get_all_items(self):
        items = []
        while not self.empty():
            items.append(self.get())
        return items

    def requeue_items(self, items):
        for item in items:
            self.add_to_queue(item)

    def remove_item(self, item_id):
        # Get all items out of queue
        current_items = self.get_all_items()
        # Create list of items that will be queued again
        requeue_items = []
        for current_item in current_items:
            if current_item.get('id') != item_id:
                requeue_items.append(current_item)
        # Remove the requested item ID from the all_items set
        lock = threading.RLock()
        lock.acquire()
        if item_id in self.all_items:
            self.all_items.remove(item_id)
        lock.release()
        # Add all requeue_items items back into the queue
        self.requeue_items(requeue_items)

    def read_all_items(self):
        # Get all items out of queue
        current_items = self.get_all_items()
        # Add all requeue_items items back into the queue
        self.requeue_items(current_items)
        # Return items list
        return current_items

    def update(self, item):
        # Ensure received item is valid
        self.__validate_item(item)
        # If it is not already in message list, add it to the list and the queue
        if item.get('id') not in self.all_items:
            self.all_items.add(item.get('id'))
            self.add_to_queue(item)
        else:
            # Get all items out of queue
            current_items = self.get_all_items()
            # Create list of items that will be queued again
            # This will not include the item requested for update
            lock = threading.RLock()
            lock.acquire()
            requeue_items = []
            for current_item in current_items:
                if current_item.get('id') != item.get('id'):
                    requeue_items.append(current_item)
                    continue
                requeue_items.append(item)
            # Add all requeue_items items back into the queue
            self.requeue_items(requeue_items)


class UnmanicDataQueues(object, metaclass=SingletonType):
    _unmanic_data_queues = {}

    def __init__(self):
        pass

    def set_unmanic_data_queues(self, unmanic_data_queues):
        self._unmanic_data_queues = unmanic_data_queues

    def get_unmanic_data_queues(self):
        return self._unmanic_data_queues


class UnmanicRunningTreads(object, metaclass=SingletonType):
    _unmanic_threads = {}

    def __init__(self):
        pass

    def set_unmanic_running_threads(self, unmanic_threads):
        self._unmanic_threads = unmanic_threads

    def get_unmanic_running_thread(self, name):
        return self._unmanic_threads.get(name)


class UIServer(threading.Thread):
    config = None
    started = False
    io_loop = None
    server = None
    app = None

    def __init__(self, unmanic_data_queues, foreman, developer):
        super(UIServer, self).__init__(name='UIServer')
        self.config = config.Config()

        self.developer = developer
        self.data_queues = unmanic_data_queues
        self.logger = unmanic_data_queues["logging"].get_logger(self.name)
        self.inotifytasks = unmanic_data_queues["inotifytasks"]
        # TODO: Move all logic out of template calling to foreman.
        #  Create methods here to handle the calls and rename to foreman
        self.foreman = foreman
        self.set_logging()
        # Add a singleton for handling the data queues for sending data to unmanic's other processes
        udq = UnmanicDataQueues()
        udq.set_unmanic_data_queues(unmanic_data_queues)
        urt = UnmanicRunningTreads()
        urt.set_unmanic_running_threads(
            {
                'foreman': foreman
            }
        )

    def _log(self, message, message2='', level="info"):
        message = common.format_message(message, message2)
        getattr(self.logger, level)(message)

    def stop(self):
        if self.started:
            self.started = False
        if self.io_loop:
            self.io_loop.add_callback(self.io_loop.stop)

    def set_logging(self):
        if self.config and self.config.get_log_path():
            # Create directory if not exists
            if not os.path.exists(self.config.get_log_path()):
                os.makedirs(self.config.get_log_path())

            # Create file handler
            log_file = os.path.join(self.config.get_log_path(), 'tornado.log')
            file_handler = logging.handlers.TimedRotatingFileHandler(log_file, when='midnight', interval=1,
                                                                     backupCount=7)
            file_handler.setLevel(logging.INFO)

            # Set tornado.access logging to file. Disable propagation of logs
            tornado_access = logging.getLogger("tornado.access")
            if self.developer:
                tornado_access.setLevel(logging.DEBUG)
            else:
                tornado_access.setLevel(logging.INFO)
            tornado_access.addHandler(file_handler)
            tornado_access.propagate = False

            # Set tornado.application logging to file. Enable propagation of logs
            tornado_application = logging.getLogger("tornado.application")
            if self.developer:
                tornado_application.setLevel(logging.DEBUG)
            else:
                tornado_application.setLevel(logging.INFO)
            tornado_application.addHandler(file_handler)
            tornado_application.propagate = True  # Send logs also to root logger (command line)

            # Set tornado.general logging to file. Enable propagation of logs
            tornado_general = logging.getLogger("tornado.general")
            if self.developer:
                tornado_general.setLevel(logging.DEBUG)
            else:
                tornado_general.setLevel(logging.INFO)
            tornado_general.addHandler(file_handler)
            tornado_general.propagate = True  # Send logs also to root logger (command line)

    def update_tornado_settings(self):
        # Check if this is a development environment or not
        if self.developer:
            tornado_settings['autoreload'] = True
            tornado_settings['serve_traceback'] = True

    def run(self):
        asyncio.set_event_loop_policy(AnyThreadEventLoopPolicy())
        self.started = True

        # Configure tornado server based on config
        self.update_tornado_settings()

        # Load the app
        self.app = self.make_web_app()

        # TODO: add support for HTTPS

        # Web Server
        self.server = HTTPServer(
            self.app,
            ssl_options=None,
        )

        try:
            self.server.listen(int(self.config.get_ui_port()))
        except socket.error as e:
            self._log("Exception when setting WebUI port {}:".format(self.config.get_ui_port()), message2=str(e),
                      level="warning")
            raise SystemExit

        self.io_loop = IOLoop().current()
        self.io_loop.start()
        self.io_loop.close(True)

        self._log("Leaving UIServer loop...")

    def make_web_app(self):
        # Start with web application routes
        from unmanic.webserver.websocket import UnmanicWebsocketHandler
        app = Application([
            (r"/unmanic/websocket", UnmanicWebsocketHandler),
            (r"/unmanic/downloads/(.*)", DownloadsHandler),
            (r"/(.*)", RedirectHandler, dict(
                url="/unmanic/ui/dashboard/"
            )),
        ], **tornado_settings)

        # Add API routes
        from unmanic.webserver.api_request_router import APIRequestRouter
        app.add_handlers(r'.*', [
            (
                PathMatches(r"/unmanic/api/.*"),
                APIRequestRouter(app)
            ),
        ])

        # Add frontend routes
        from unmanic.webserver.main import MainUIRequestHandler
        app.add_handlers(r'.*', [
            (r"/unmanic/css/(.*)", StaticFileHandler, dict(
                path=tornado_settings['static_css']
            )),
            (r"/unmanic/fonts/(.*)", StaticFileHandler, dict(
                path=tornado_settings['static_fonts']
            )),
            (r"/unmanic/icons/(.*)", StaticFileHandler, dict(
                path=tornado_settings['static_icons']
            )),
            (r"/unmanic/img/(.*)", StaticFileHandler, dict(
                path=tornado_settings['static_img']
            )),
            (r"/unmanic/js/(.*)", StaticFileHandler, dict(
                path=tornado_settings['static_js']
            )),
            (
                PathMatches(r"/unmanic/ui/(.*)"),
                MainUIRequestHandler,
            ),
        ])

        # Add widgets routes
        from unmanic.webserver.plugins import DataPanelRequestHandler
        from unmanic.webserver.plugins import PluginStaticFileHandler
        from unmanic.webserver.plugins import PluginAPIRequestHandler
        app.add_handlers(r'.*', [
            (
                PathMatches(r"/unmanic/panel/[^/]+(/(?!static/|assets$).*)?$"),
                DataPanelRequestHandler
            ),
            (
                PathMatches(r"/unmanic/plugin_api/[^/]+(/(?!static/|assets$).*)?$"),
                PluginAPIRequestHandler
            ),
            (r"/unmanic/panel/.*/static/(.*)", PluginStaticFileHandler, dict(
                path=tornado_settings['static_img']
            )),
        ])

        if self.developer:
            self._log("API Docs - Updating...", level="debug")
            try:
                from unmanic.webserver.api_v2.schema.swagger import generate_swagger_file
                errors = generate_swagger_file()
                for error in errors:
                    self._log(error, level="warn")
                else:
                    self._log("API Docs - Updated successfully", level="debug")
            except Exception as e:
                self._log("Failed to reload API schema", message2=str(e), level="error")

        # Start the Swagger UI. Automatically generated swagger.json can also
        # be served using a separate Swagger-service.
        from swagger_ui import tornado_api_doc
        tornado_api_doc(
            app,
            config_path=os.path.join(os.path.dirname(__file__), "..", "webserver", "docs", "api_schema_v2.json"),
            url_prefix="/unmanic/swagger",
            title="Unmanic application API"
        )

        return app



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.taskqueue.py

    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     23 Apr 2019, (19:17 PM)

    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved

           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:

           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.

           THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""

from unmanic.libs import task
from unmanic.libs import common
from unmanic.libs.unmodels import Libraries, LibraryTags, Tags
from unmanic.libs.unmodels.tasks import Tasks

"""

An object to contain all details of the job queue in such a way that it is presented in a synchronous list
while being able to be accessed by a number of threads simultaneously

"""


def build_tasks_count_query(status):
    """
    Return a 0 if no tasks exist for the given status.
    Return a count >= 1 if any tasks exist for the given status.

    # TODO: look into peewee dynamic query building (surly this exists)

    :param status:
    :return:
    """
    # Fetch only on result in order to know that there are any at all
    # Filter by status
    query = Tasks.select().where((Tasks.status == status)).limit(1)
    return query.count()


def build_tasks_query(status, sort_by='id', sort_order='asc', local_only=False, library_names=None, library_tags=None):
    """
    Return the first task item in the task list filtered by status
    and sorted by the self.sort_by and self.sort_order variables.

    :param status:
    :param sort_order:
    :param sort_by:
    :param local_only:
    :param library_names:
    :param library_tags:
    :return:
    """
    # pick query based on sort params
    query = Tasks.select().where((Tasks.status == status))

    # Limit to one result
    if local_only:
        query = query.where((Tasks.type == 'local'))

    query = query.join(Libraries, on=(Libraries.id == Tasks.library_id))
    if library_names is not None:
        query = query.where(Libraries.name.in_(library_names))
    if library_tags is not None:
        query = query.join(LibraryTags, join_type='LEFT OUTER JOIN')
        query = query.join(Tags, join_type='LEFT OUTER JOIN')
        if library_tags:
            query = query.where(Tags.name.in_(library_tags))
        else:
            # Handle a query where the list is empty. In this case we want to match for only libraries that have no tags
            query = query.where(Tags.name.is_null())

    # Limit to one result
    query = query.limit(1)
    if sort_order == 'asc':
        query = query.order_by(sort_by.asc())
    else:
        query = query.order_by(sort_by.desc())
    return query.first()


def build_tasks_query_full_task_list(status, sort_by='id', sort_order='asc', limit=None):
    """
    Return all task items in the task list filtered by status.
    The query is sorted by the self.sort_by and self.sort_order variables
    and may be limited by the limit variable.

    :param sort_order:
    :param sort_by:
    :param status:
    :param limit:
    :return:
    """
    query = Tasks.select(Tasks).where((Tasks.status == status))

    # Set the sort order
    if sort_order == 'asc':
        query = query.order_by(sort_by.asc())
    else:
        query = query.order_by(sort_by.desc())

    # Set query limit if one was given
    if limit:
        query = query.limit(limit)

    # Return results as dictionary
    return query.dicts()


def fetch_next_task_filtered(status, sort_by='id', sort_order='asc', local_only=False, library_names=None, library_tags=None):
    """
    Returns the next task in the task list for a given status

    :param status:
    :param sort_order:
    :param sort_by:
    :param local_only:
    :param library_names:
    :param library_tags:
    :return:
    """
    # Fetch the task item first (to ensure it exists)
    task_item = build_tasks_query(status, sort_by=sort_by, sort_order=sort_order, local_only=local_only,
                                  library_names=library_names, library_tags=library_tags)
    if not task_item:
        return False
    # Set the task object by the abspath and return it
    next_task = task.Task()
    next_task.read_and_set_task_by_absolute_path(task_item.abspath)
    return next_task


class TaskQueue(object):
    """
    TaskQueue

    Creates an job item per file.
    This job item is passed through stages by the Foreman and PostProcessor

    Attributes:
        data_queues (list): A list of Queue objects. Contains the logger

    """

    def __init__(self, data_queues):
        self.name = 'TaskQueue'
        self.data_queues = data_queues
        self.logger = data_queues["logging"].get_logger(self.name)

        # Sort fields
        self.sort_by = Tasks.priority
        self.sort_order = 'desc'

    def _log(self, message, message2='', level="info"):
        message = common.format_message(message, message2)
        getattr(self.logger, level)(message)

    """
    Last task based on status pending, in_progress or processed
    """

    def list_pending_tasks(self, limit=None):
        """
        Returns a list of 'pending' tasks
        Can limit to <limit> results

        :param limit:
        :return:
        """
        results = build_tasks_query_full_task_list('pending', self.sort_by, self.sort_order, limit)
        if results:
            return list(results)
        return []

    def list_in_progress_tasks(self, limit=None):
        """
        Returns a list of 'in_progress' tasks
        Can limit to <limit> results

        :param limit:
        :return:
        """
        results = build_tasks_query_full_task_list('in_progress', self.sort_by, self.sort_order, limit)
        if results:
            return list(results)
        return []

    def list_processed_tasks(self, limit=None):
        """
        Returns a list of 'processed' tasks
        Can limit to <limit> results

        :param limit:
        :return:
        """
        results = build_tasks_query_full_task_list('processed', self.sort_by, self.sort_order, limit)
        if results:
            return list(results)
        return []

    """
    Get first task in task list based on status pending, in_progress or processed
    """

    def get_next_pending_tasks(self, local_only=False, library_names=None, library_tags=None):
        """
        Fetch the next pending task.
        Set that task status as 'in_progress' and then return it.

        :param local_only:
        :param library_names:
        :param library_tags:
        :return:
        """
        # Fetch Task item matching the filters specified
        task_item = fetch_next_task_filtered('pending', sort_by=self.sort_by, sort_order=self.sort_order,
                                             local_only=local_only, library_names=library_names, library_tags=library_tags)
        return task_item

    def get_next_processed_tasks(self):
        # Fetch Task item matching the filters specified
        task_item = fetch_next_task_filtered('processed', sort_by=self.sort_by, sort_order=self.sort_order)
        return task_item

    def requeue_tasks_at_bottom(self, task_id):
        task_handler = task.Task()
        return task_handler.reorder_tasks([task_id], 'bottom')

    """
    Check if a particular task list is empty
    """

    @staticmethod
    def task_list_pending_is_empty():
        # Fetch only on result in order to know that there are any at all
        pending_query_count = build_tasks_count_query('pending')
        if pending_query_count > 0:
            return False
        return True

    @staticmethod
    def task_list_in_progress_is_empty():
        # Fetch only on result in order to know that there are any at all
        pending_query_count = build_tasks_count_query('in_progress')
        if pending_query_count > 0:
            return False
        return True

    @staticmethod
    def task_list_processed_is_empty():
        # Fetch only on result in order to know that there are any at all
        pending_query_count = build_tasks_count_query('processed')
        if pending_query_count > 0:
            return False
        return True

    """
    Set the status of a task item
    """

    @staticmethod
    def mark_item_in_progress(task_item):
        """
        Set the given task status as 'in_progress' and then return it.

        :param task_item:
        :return:
        """
        # Set item as status = 'in_progress'
        task_item.set_status('in_progress')
        return task_item

    @staticmethod
    def mark_item_as_processed(task_item):
        """
        Set the given task status as 'processed' and then return it.

        :param task_item:
        :return:
        """
        # Set item as status = 'processed'
        task_item.set_status('processed')
        return task_item



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.plugins_api.py

    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     01 Aug 2021, (9:30 AM)

    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved

           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:

           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.

           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""

import tornado.log
from unmanic.libs import session
from unmanic.libs.uiserver import UnmanicDataQueues
from unmanic.webserver.api_v2.base_api_handler import BaseApiHandler, BaseApiError
from unmanic.webserver.api_v2.schema.schemas import PluginFlowResultsSchema, PluginReposListResultsSchema, \
    PluginTypesResultsSchema, PluginsDataPanelTypesDataSchema, PluginsDataSchema, PluginsInfoResultsSchema, \
    PluginsInstallableResultsSchema, RequestPluginsByIdSchema, RequestPluginsFlowByPluginTypeSchema, \
    RequestPluginsInfoSchema, RequestPluginsSettingsResetSchema, RequestPluginsSettingsSaveSchema, \
    RequestPluginsTableDataSchema, \
    RequestSavingPluginsFlowByPluginTypeSchema, RequestTableUpdateByIdList, RequestUpdatePluginReposListSchema
from unmanic.webserver.helpers import plugins


class ApiPluginsHandler(BaseApiHandler):
    session = None
    config = None
    params = None
    unmanic_data_queues = None

    routes = [
        {
            "path_pattern":      r"/plugins/installed",
            "supported_methods": ["POST"],
            "call_method":       "get_installed_plugins",
        },
        {
            "path_pattern":      r"/plugins/enable",
            "supported_methods": ["POST"],
            "call_method":       "enable_plugins",
        },
        {
            "path_pattern":      r"/plugins/disable",
            "supported_methods": ["POST"],
            "call_method":       "disable_plugins",
        },
        {
            "path_pattern":      r"/plugins/update",
            "supported_methods": ["POST"],
            "call_method":       "update_plugins",
        },
        {
            "path_pattern":      r"/plugins/remove",
            "supported_methods": ["DELETE"],
            "call_method":       "remove_plugins",
        },
        {
            "path_pattern":      r"/plugins/info",
            "supported_methods": ["POST"],
            "call_method":       "get_plugin_info",
        },
        {
            "path_pattern":      r"/plugins/settings/update",
            "supported_methods": ["POST"],
            "call_method":       "update_plugin_settings",
        },
        {
            "path_pattern":      r"/plugins/settings/reset",
            "supported_methods": ["POST"],
            "call_method":       "reset_plugin_settings",
        },
        {
            "path_pattern":      r"/plugins/installable",
            "supported_methods": ["GET"],
            "call_method":       "get_installable_plugin_list",
        },
        {
            "path_pattern":      r"/plugins/install",
            "supported_methods": ["POST"],
            "call_method":       "install_plugin_by_id",
        },
        {
            "path_pattern":      r"/plugins/flow",
            "supported_methods": ["POST"],
            "call_method":       "get_enabled_plugins_flow_by_type",
        },
        {
            "path_pattern":      r"/plugins/flow/types",
            "supported_methods": ["GET"],
            "call_method":       "get_plugin_types_with_flows",
        },
        {
            "path_pattern":      r"/plugins/flow/save",
            "supported_methods": ["POST"],
            "call_method":       "save_enabled_plugin_flow",
        },
        {
            "path_pattern":      r"/plugins/repos/update",
            "supported_methods": ["POST"],
            "call_method":       "update_repo_list",
        },
        {
            "path_pattern":      r"/plugins/repos/list",
            "supported_methods": ["GET"],
            "call_method":       "get_repo_list",
        },
        {
            "path_pattern":      r"/plugins/repos/reload",
            "supported_methods": ["POST"],
            "call_method":       "reload_repo_data",
        },
        {
            "path_pattern":      r"/plugins/panels/enabled",
            "supported_methods": ["GET"],
            "call_method":       "get_enabled_panel_plugins_list",
        },
    ]

    def initialize(self, **kwargs):
        self.session = session.Session()
        self.params = kwargs.get("params")
        udq = UnmanicDataQueues()
        self.unmanic_data_queues = udq.get_unmanic_data_queues()

    def get_installed_plugins(self):
        """
        Plugins - list installed plugins
        ---
        description: Returns a list of installed plugins.
        requestBody:
            description: Requested a list of installed plugins.
            required: True
            content:
                application/json:
                    schema:
                        RequestPluginsTableDataSchema
        responses:
            200:
                description: 'Sample response: Returns a list of installed plugins.'
                content:
                    application/json:
                        schema:
                            PluginsDataSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            json_request = self.read_json_request(RequestPluginsTableDataSchema())

            params = {
                'start':        json_request.get('start', '0'),
                'length':       json_request.get('length', '10'),
                'search_value': json_request.get('search_value', ''),
                'order':        {
                    "column": json_request.get('order_by', 'name'),
                    "dir":    json_request.get('order_direction', 'asc'),
                }
            }
            plugins_list = plugins.prepare_filtered_plugins(params)

            response = self.build_response(
                PluginsDataSchema(),
                {
                    "recordsTotal":    plugins_list.get('recordsTotal'),
                    "recordsFiltered": plugins_list.get('recordsFiltered'),
                    "results":         plugins_list.get('results'),
                }
            )
            self.write_success(response)
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def enable_plugins(self):
        """
        Plugins - enable
        ---
        description: DEPRECATED! Enable a list of plugins.
        requestBody:
            description: Requested list of plugins to enable.
            required: True
            content:
                application/json:
                    schema:
                        RequestTableUpdateByIdList
        responses:
            200:
                description: 'Successful request; Returns success status'
                content:
                    application/json:
                        schema:
                            BaseSuccessSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            raise Exception('Endpoint is deprecated. Plugins are now enabled by assigning them to a library')
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def disable_plugins(self):
        """
        Plugins - disable
        ---
        description: DEPRECATED! Disable a list of plugins.
        requestBody:
            description: Requested list of plugins to disable.
            required: True
            content:
                application/json:
                    schema:
                        RequestTableUpdateByIdList
        responses:
            200:
                description: 'Successful request; Returns success status'
                content:
                    application/json:
                        schema:
                            BaseSuccessSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            raise Exception('Endpoint is deprecated. Plugins are now enabled by assigning them to a library')
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def update_plugins(self):
        """
        Plugins - update
        ---
        description: Update a list of plugins given their DB table IDs.
        requestBody:
            description: Requested list of plugins to update.
            required: True
            content:
                application/json:
                    schema:
                        RequestTableUpdateByIdList
        responses:
            200:
                description: 'Successful request; Returns success status'
                content:
                    application/json:
                        schema:
                            BaseSuccessSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            json_request = self.read_json_request(RequestTableUpdateByIdList())

            if not plugins.update_plugins(json_request.get('id_list', [])):
                self.set_status(self.STATUS_ERROR_INTERNAL, reason="Failed to update the plugins by their IDs")
                self.write_error()
                return

            self.write_success()
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def remove_plugins(self):
        """
        Plugins - remove
        ---
        description: Remove a list of plugins given their DB table IDs.
        requestBody:
            description: Requested list of plugins to remove.
            required: True
            content:
                application/json:
                    schema:
                        RequestTableUpdateByIdList
        responses:
            200:
                description: 'Successful request; Returns success status'
                content:
                    application/json:
                        schema:
                            BaseSuccessSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            json_request = self.read_json_request(RequestTableUpdateByIdList())

            if not plugins.remove_plugins(json_request.get('id_list', [])):
                self.set_status(self.STATUS_ERROR_INTERNAL, reason="Failed to remove the plugins by their IDs")
                self.write_error()
                return

            self.write_success()
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def get_plugin_info(self):
        """
        Plugins - return a requested plugin's metadata and settings
        ---
        description: Returns a the metadata and settings of a requested plugin.
        requestBody:
            description: Requested a single plugin's info.
            required: True
            content:
                application/json:
                    schema:
                        RequestPluginsInfoSchema
        responses:
            200:
                description: 'Sample response: Returns a the metadata and settings of a requested plugin.'
                content:
                    application/json:
                        schema:
                            PluginsInfoResultsSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            json_request = self.read_json_request(RequestPluginsInfoSchema())

            plugin_id = json_request.get('plugin_id')
            prefer_local = json_request.get('prefer_local')
            library_id = json_request.get('library_id')

            plugins_info = plugins.prepare_plugin_info_and_settings(plugin_id,
                                                                    prefer_local=prefer_local,
                                                                    library_id=library_id)

            response = self.build_response(
                PluginsInfoResultsSchema(),
                {
                    "plugin_id":   plugins_info.get('plugin_id'),
                    "icon":        plugins_info.get('icon'),
                    "name":        plugins_info.get('name'),
                    "description": plugins_info.get('description'),
                    "tags":        plugins_info.get('tags'),
                    "author":      plugins_info.get('author'),
                    "version":     plugins_info.get('version'),
                    "changelog":   plugins_info.get('changelog'),
                    "status":      plugins_info.get('status'),
                    "settings":    plugins_info.get('settings'),
                }
            )
            self.write_success(response)
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def update_plugin_settings(self):
        """
        Plugins - Save the settings of a single plugin
        ---
        description: Saves the settings of a single plugin.
        requestBody:
            description: Requested a plugins settings be updated.
            required: True
            content:
                application/json:
                    schema:
                        RequestPluginsSettingsSaveSchema
        responses:
            200:
                description: 'Successful request; Returns success status'
                content:
                    application/json:
                        schema:
                            BaseSuccessSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            json_request = self.read_json_request(RequestPluginsSettingsSaveSchema())

            plugin_id = json_request.get('plugin_id')
            settings = json_request.get('settings')
            library_id = json_request.get('library_id')

            if not plugins.update_plugin_settings(plugin_id, settings, library_id=library_id):
                self.set_status(self.STATUS_ERROR_INTERNAL, reason="Failed to save plugins settings")
                self.write_error()
                return

            self.write_success()
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def reset_plugin_settings(self):
        """
        Plugins - Reset the settings of a single plugin
        ---
        description: Reset the settings of a single plugin.
        requestBody:
            description: Requested a plugins settings be reset.
            required: True
            content:
                application/json:
                    schema:
                        RequestPluginsSettingsResetSchema
        responses:
            200:
                description: 'Successful request; Returns success status'
                content:
                    application/json:
                        schema:
                            BaseSuccessSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            json_request = self.read_json_request(RequestPluginsSettingsResetSchema())

            plugin_id = json_request.get('plugin_id')
            library_id = json_request.get('library_id')

            if not plugins.reset_plugin_settings(plugin_id, library_id=library_id):
                self.set_status(self.STATUS_ERROR_INTERNAL, reason="Failed to reset plugins settings")
                self.write_error()
                return

            self.write_success()
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def get_installable_plugin_list(self):
        """
        Plugins - Read all installable plugins
        ---
        description: Returns a list of installable plugins.
        responses:
            200:
                description: 'Sample response: Returns a list of installable plugins.'
                content:
                    application/json:
                        schema:
                            PluginsInstallableResultsSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            installable_plugins_list = plugins.prepare_installable_plugins_list()

            response = self.build_response(
                PluginsInstallableResultsSchema(),
                {
                    "plugins": installable_plugins_list
                }
            )
            self.write_success(response)
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def install_plugin_by_id(self):
        """
        Plugins - Install a single plugin by its Plugin ID
        ---
        description: Installs a plugin by its Plugin ID.
        requestBody:
            description: Requested a plugin be installed by its Plugin ID.
            required: True
            content:
                application/json:
                    schema:
                        RequestPluginsByIdSchema
        responses:
            200:
                description: 'Successful request; Returns success status'
                content:
                    application/json:
                        schema:
                            BaseSuccessSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            json_request = self.read_json_request(RequestPluginsByIdSchema())

            if not plugins.install_plugin_by_id(json_request.get('plugin_id'), json_request.get('repo_id')):
                self.set_status(self.STATUS_ERROR_INTERNAL, reason="Failed to install/update plugin")
                self.write_error()
                return

            self.write_success()
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def get_plugin_types_with_flows(self):
        """
        Plugins - Get a list of all plugin types that have flows
        ---
        description: Returns a list of all plugin types that have flows.
        responses:
            200:
                description: 'Sample response: Returns a list of all plugin types that have flows.'
                content:
                    application/json:
                        schema:
                            PluginTypesResultsSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            results = plugins.get_plugin_types_with_flows()
            response = self.build_response(
                PluginTypesResultsSchema(),
                {
                    "results": results,
                }
            )
            self.write_success(response)
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def get_enabled_plugins_flow_by_type(self):
        """
        Plugins - Get the plugin flow for a requested plugin type
        ---
        description: Returns the plugin flow for a requested plugin type.
        requestBody:
            description: Requests the plugin flow of a given plugin type.
            required: True
            content:
                application/json:
                    schema:
                        RequestPluginsFlowByPluginTypeSchema
        responses:
            200:
                description: 'Sample response: Returns the plugin flow for a requested plugin type.'
                content:
                    application/json:
                        schema:
                            PluginsInstallableResultsSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            json_request = self.read_json_request(RequestPluginsFlowByPluginTypeSchema())

            results = plugins.get_enabled_plugin_flows_for_plugin_type(json_request.get('plugin_type'),
                                                                       json_request.get('library_id'))
            response = self.build_response(
                PluginFlowResultsSchema(),
                {
                    "results": results,
                }
            )
            self.write_success(response)
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def save_enabled_plugin_flow(self):
        """
        Plugins - Save the plugin flow for a requested plugin type
        ---
        description: Saves the plugin flow for a requested plugin type.
        requestBody:
            description: Requests saving the plugin flow for a given plugin type.
            required: True
            content:
                application/json:
                    schema:
                        RequestSavingPluginsFlowByPluginTypeSchema
        responses:
            200:
                description: 'Sample response: Save the plugin flow for a requested plugin type.'
                content:
                    application/json:
                        schema:
                            PluginsInstallableResultsSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            json_request = self.read_json_request(RequestSavingPluginsFlowByPluginTypeSchema())

            if not plugins.save_enabled_plugin_flows_for_plugin_type(json_request.get('plugin_type'),
                                                                     json_request.get('library_id'),
                                                                     json_request.get('plugin_flow')):
                self.set_status(self.STATUS_ERROR_INTERNAL, reason="Failed to update plugin flow by type")
                self.write_error()
                return

            self.write_success()
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def update_repo_list(self):
        """
        Plugins - Update the plugin repo list
        ---
        description: Updates the plugin repo list.
        requestBody:
            description: Requested an update to the plugin repo list.
            required: True
            content:
                application/json:
                    schema:
                        RequestUpdatePluginReposListSchema
        responses:
            200:
                description: 'Successful request; Returns success status'
                content:
                    application/json:
                        schema:
                            BaseSuccessSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            json_request = self.read_json_request(RequestUpdatePluginReposListSchema())

            if not plugins.save_plugin_repos_list(json_request.get('repos_list')):
                self.set_status(self.STATUS_ERROR_INTERNAL, reason="Failed to update plugin repo list")
                self.write_error()
                return

            self.write_success()
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def get_repo_list(self):
        """
        Plugins - Read all configured plugin repos
        ---
        description: Returns a list of plugin repos.
        responses:
            200:
                description: 'Sample response: Returns a list of plugin repos.'
                content:
                    application/json:
                        schema:
                            PluginReposListResultsSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            plugin_repos_list = plugins.prepare_plugin_repos_list()

            response = self.build_response(
                PluginReposListResultsSchema(),
                {
                    "repos": plugin_repos_list
                }
            )
            self.write_success(response)
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def reload_repo_data(self):
        """
        Plugins - Reload plugin repositories remote data
        ---
        description: Reload plugin repositories remote data.
        responses:
            200:
                description: 'Successful request; Returns success status'
                content:
                    application/json:
                        schema:
                            BaseSuccessSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            if not plugins.reload_plugin_repos_data():
                self.set_status(self.STATUS_ERROR_INTERNAL, reason="Failed to pull latest plugin repo data")
                self.write_error()
                return

            self.write_success()
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def get_enabled_panel_plugins_list(self):
        """
        Plugins - Read all enabled "data panel" type plugins
        ---
        description: Returns a list of installed plugins.
        responses:
            200:
                description: 'Success: Returns a list of installed plugins.'
                content:
                    application/json:
                        schema:
                            PluginsDataPanelTypesDataSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            data_panel_plugins = plugins.get_enabled_plugin_data_panels()

            # Only return the data that we need
            plugin_list = []
            for data_panel_plugin in data_panel_plugins:
                plugin_list.append(
                    {
                        "plugin_id":   data_panel_plugin.get("plugin_id"),
                        "name":        data_panel_plugin.get("name", ""),
                        "author":      data_panel_plugin.get("author", ""),
                        "description": data_panel_plugin.get("description", ""),
                        "version":     data_panel_plugin.get("version", ""),
                        "icon":        data_panel_plugin.get("icon", ""),
                    }
                )

            response = self.build_response(
                PluginsDataPanelTypesDataSchema(),
                {
                    "results": plugin_list
                }
            )
            self.write_success(response)
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.settings.py

    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     16 Mar 2021, (7:14 PM)

    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved

           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:

           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.

           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""
import json
import os
import sys

from unmanic import config
from unmanic.libs.singleton import SingletonType


class PluginSettings(object):
    """
    A dictionary of settings accessible to the Plugin class and able
    to be configured by users from within the Unmanic WebUI.

    """
    settings = {}

    """
    A dictionary of form settings used by Unmanic's WebUI to configure
    the plugin's settings form.

    """
    form_settings = {}

    """
    A cached copy of settings as they are stored on disk.

    """
    settings_configured = None

    """
    The library ID that we are fetching settings for.
    
    """
    library_id = None

    def __init__(self, *args, **kwargs):
        self.library_id = kwargs.get('library_id')
        # If the given library is not None, ensure that it is a number
        if self.library_id:
            try:
                self.library_id = int(self.library_id)
            except ValueError:
                raise Exception("Library ID needs to be an integer. You have provided '{}'".format(self.library_id))

    def __get_plugin_settings_file(self, force_library_settings=False):
        plugin_directory = self.get_plugin_directory()
        profile_directory = self.get_profile_directory()
        # Temp code to migrate settings to userdata
        # TODO: Remove after initial release
        if not os.path.exists(os.path.join(profile_directory, 'settings.json')):
            if os.path.exists(os.path.join(plugin_directory, 'settings.json')):
                import shutil
                shutil.move(
                    os.path.join(plugin_directory, 'settings.json'),
                    os.path.join(profile_directory, 'settings.json')
                )
        # If provided with a library ID, then the settings file will be different
        plugin_settings_file = os.path.join(profile_directory, 'settings.json')
        if self.library_id:
            plugin_settings_file = os.path.join(profile_directory, 'settings.{}.json'.format(self.library_id))
            if not os.path.exists(plugin_settings_file) and not force_library_settings:
                # If the library file does not yet exist, then resort to using the default settings file
                plugin_settings_file = os.path.join(profile_directory, 'settings.json')
        return plugin_settings_file

    def __export_configured_settings(self):
        """
        Write settings to settings file

        :return:
        """
        plugin_settings_file = self.__get_plugin_settings_file(force_library_settings=True)

        with open(plugin_settings_file, 'w') as f:
            json.dump(self.settings_configured, f, indent=2)

    def __import_configured_settings(self):
        """
        Read settings from settings file

        :return:
        """
        plugin_settings_file = self.__get_plugin_settings_file()

        # Default the configured settings to the plugin defaults
        # Loop over the self.settings object to clone the keys/values
        self.settings_configured = {}
        for key in self.settings:
            self.settings_configured[key] = self.settings[key]

        # if the file does not yet exist, create it
        if not os.path.exists(plugin_settings_file):
            self.__export_configured_settings()

        # Read plugin settings from file
        with open(plugin_settings_file) as infile:
            plugin_settings = json.load(infile)

        # Loop over settings
        for key in self.settings:
            if key in plugin_settings:
                self.settings_configured[key] = plugin_settings.get(key)

    def reset_settings_to_defaults(self):
        """
        Remove all currently configured settings by deleting the settings.json file

        :return:
        """
        plugin_settings_file = self.__get_plugin_settings_file()

        # If the settings file returned is the global settings file and this was called on a library config,
        # do not reset the config.
        if self.library_id is not None and os.path.basename(plugin_settings_file) == 'settings.json':
            return False

        # if the file does not yet exist, create it
        if os.path.exists(plugin_settings_file):
            os.remove(plugin_settings_file)

        if not os.path.exists(plugin_settings_file):
            return True
        return False

    def get_plugin_directory(self):
        """
        Return the absolute path to the Plugin's directory.
        This is where the Plugin is currently installed.

        :return:
        """
        return os.path.dirname(os.path.abspath(sys.modules[self.__class__.__module__].__file__))

    def get_profile_directory(self):
        """
        Return the absolute path to the Plugin's profile directory.
        This is where where Plugin settings are saved and where all mutable data for the
        Plugin should be stored.

        :return:
        """
        settings = config.Config()
        userdata_path = settings.get_userdata_path()
        plugin_directory = self.get_plugin_directory()
        plugin_id = os.path.basename(plugin_directory)
        profile_directory = os.path.join(userdata_path, plugin_id)
        if not os.path.exists(profile_directory):
            os.makedirs(profile_directory)
        return profile_directory

    def get_form_settings(self):
        """
        Return the current form settings.

        :return:
        """
        return self.form_settings

    def get_setting(self, key=None):
        """
        Fetch a single configuration value, or, when passed "all" as the key argument,
        return the full configuration dictionary.

        :param key:
        :return:
        """
        # First import settings
        try:
            self.__import_configured_settings()
        except json.decoder.JSONDecodeError:
            # If the import fails, then it will resort to defaults.
            # That is fine. Better than breaking the rest of the process
            pass
        except FileNotFoundError:
            # If the settings file did not exist, then also resort to defaults.
            pass

        if key is None:
            return self.settings_configured
        else:
            return self.settings_configured.get(key)

    def set_setting(self, key, value):
        """
        Set a singe configuration value.
        Used by the Unmanic WebUI to save user settings.
        Settings are stored on disk in order to be persistent.

        :param key:
        :param value:
        :return:
        """
        # First import settings
        try:
            self.__import_configured_settings()
        except json.decoder.JSONDecodeError:
            # If the import fails, then it will resort to defaults.
            # That is fine. Better than breaking the rest of the process
            pass

        # Ensure plugin has this setting
        if key not in self.settings:
            return False

        # Set the configured value
        self.settings_configured[key] = value

        # Export the settings again
        self.__export_configured_settings()

        return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.directoryinfo.py

    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     02 Jul 2021, (10:59 AM)

    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved

           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:

           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.

           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""
import json
import os
import configparser


class UnmanicDirectoryInfoException(Exception):
    def __init__(self, message, path):
        errmsg = '%s: file %s' % (message, path)
        Exception.__init__(self, errmsg)
        self.message = message
        self.path = path

    def __repr__(self):
        return self.message

    __str__ = __repr__


class UnmanicDirectoryInfo:
    """
    UnmanicDirectoryInfo

    Manages the reading and writing of the '.unmanic' files located in the directories
    parsed by Unmanic's library scanner or any plugins.

    Legacy support:
        On read, if the config is an INI file, uses ConfigParser.get to fetch information
        and convert it to JSON.
        INI was initially used in order to be a simple syntax for manually editing.
        However, INI is not ideal for managing file names and paths with special characters

    """

    def __init__(self, directory):
        self.path = os.path.join(directory, '.unmanic')
        self.json_data = None
        self.config_parser = None
        # If the path does not exist, do not try to read it
        if not os.path.exists(self.path):
            self.json_data = {}
            return
        # First read JSON data
        try:
            with open(self.path) as infile:
                self.json_data = json.load(infile)
            # Migrate JSON to latest formatting
            self.__migrate_json_formatting()
        except json.decoder.JSONDecodeError:
            pass
        # If we were unable to import the JSON data, attempt to read as INI
        if self.json_data is None:
            try:
                self.config_parser = configparser.ConfigParser(allow_no_value=True)
                self.config_parser.read(self.path)
                # Migrate file to JSON
                self.__migrate_to_json()
            except configparser.MissingSectionHeaderError:
                pass
            except configparser.NoSectionError:
                pass
            except configparser.NoOptionError:
                pass
        # If we still do not have JSON data at this point, something has gone wrong
        if self.json_data is None:
            raise UnmanicDirectoryInfoException("Failed to read directory info", self.path)

    def __migrate_to_json(self):
        """
        Migrate data from INI to JSON

        :return:
        """
        sections = self.config_parser.sections()
        json_data = {}
        for section in sections:
            section_data = {}
            for key in self.config_parser[section]:
                section_data[key.lower()] = self.config_parser.get(section, key)
            json_data[section] = section_data
        self.json_data = json_data

    def __migrate_json_formatting(self):
        """
        Migrate JSON to latest format

        Migration 1:
            As Unmanic may be used on platforms that view files as case-insensitive,
            we should ensure that all keys are also stored this way.

        :return:
        """
        # Ensure all keys are lower case
        sections = [s for s in self.json_data]
        for section in sections:
            # Sections remain case sensitive, but keys must be lowercase
            keys = [k for k in self.json_data[section]]
            for key in keys:
                if key != key.lower():
                    self.json_data[section][key.lower()] = self.json_data[section][key]
                    del self.json_data[section][key]

    def set(self, section, option, value=None):
        """
        Set an option.

        :param section:
        :param option:
        :param value:
        :return:
        """
        # Ensure keys are always lower-case
        option = option.lower()
        if self.json_data is not None:
            if not self.json_data.get(section):
                self.json_data[section] = {}
            self.json_data[section][option] = value
            return
        elif self.config_parser:
            if not self.config_parser.has_section(section):
                self.config_parser.add_section(section)
            self.config_parser.set(section, option, value)
            return
        raise UnmanicDirectoryInfoException("Failed to set section '{}' option '{}' value '{}'".format(section, option, value),
                                            self.path)

    def get(self, section, option):
        """
        Get an option

        :param section:
        :param option:
        :return:
        """
        option = option.lower()
        if self.json_data is not None:
            return self.json_data.get(section, {}).get(option)
        elif self.config_parser:
            return self.config_parser.get(section, option)
        raise UnmanicDirectoryInfoException("Failed to get section '{}' option '{}'".format(section, option), self.path)

    def save(self):
        """
        Saves the data to file.

        :return:
        """
        if self.json_data is not None:
            with open(self.path, 'w') as outfile:
                json.dump(self.json_data, outfile, indent=2)
            return
        elif self.config_parser:
            with open(self.path, 'w') as outfile:
                self.config_parser.write(outfile)
            return
        raise UnmanicDirectoryInfoException("Failed to save directory info", self.path)


if __name__ == '__main__':
    directory_info = UnmanicDirectoryInfo('/tmp/unmanic')
    directory_info.set('test_section', 'key', 'value')
    directory_info.save()
    print(directory_info.get('test_section', 'key'))
    directory_info.set('"section with double quotes"', '"key with double quotes"', '"value with double quotes"')
    directory_info.save()
    print(directory_info.get('"section with double quotes"', '"key with double quotes"'))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.setup.py
 
    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     04 May 2020, (10:47 AM)
 
    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved
 
           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:
  
           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.
  
           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""
import json
import os
import shutil
import subprocess
import sys
import glob
from setuptools import setup, find_packages, Command
import setuptools.command.build_py

if sys.version_info[0] < 3:
    print("This module version requires Python 3.")
    sys.exit(1)

sys.path.insert(0, os.path.join(os.path.dirname(__file__)))
import versioninfo

project_root_dir = os.path.dirname(os.path.realpath(__file__))
src_dir = 'unmanic'

module_name = versioninfo.name()
module_version = versioninfo.version()
module_description = versioninfo.description()
module_author = versioninfo.author()
module_email = versioninfo.email()
module_url = versioninfo.url()
module_classifiers = [
    versioninfo.dev_status(),
    'Intended Audience :: End Users/Desktop',
    'Intended Audience :: Developers',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Operating System :: POSIX :: Linux',
    'Operating System :: Unix',
    'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    'Topic :: Multimedia :: Video :: Conversion',
    'Topic :: Internet :: WWW/HTTP',
]


class BuildPyCommand(setuptools.command.build_py.build_py):
    """Custom build command."""

    def run(self):
        setuptools.command.build_py.build_py.run(self)
        self.run_command('write-build-version')
        self.run_command('build-frontend')


class WriteVersionCommand(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    @staticmethod
    def run():
        version_file = os.path.join('.', 'build', 'lib', src_dir, 'version')
        data = {
            'short': versioninfo.version(),
            'long':  versioninfo.full_version(),
        }
        with open(version_file, 'w') as f:
            json.dump(data, f)


class BuildFrontendCommand(setuptools.command.build_py.build_py):
    """Frontend build command."""

    def run(self):
        setuptools.command.build_py.build_py.run(self)

        public_asset_path = os.path.abspath(os.path.join('.', 'build', 'lib', src_dir, 'webserver', 'public'))
        frontend_path = os.path.abspath(os.path.join('.', 'build', 'lib', src_dir, 'webserver', 'frontend'))

        # Start by clearing out anything if this was pulled from a dirty tree
        shutil.rmtree(public_asset_path, ignore_errors=True)
        shutil.rmtree(os.path.join(frontend_path, 'node_modules'), ignore_errors=True)

        # Install all modules
        subprocess.run(
            "npm ci",
            check=True,
            shell=True,
            cwd=frontend_path,
        )
        # Build the frontend
        subprocess.run(
            "npm run build:publish",
            check=True,
            shell=True,
            cwd=frontend_path,
        )

        # Move built dist to templates directory
        shutil.move(os.path.join(frontend_path, 'dist', 'spa'), public_asset_path)
        # Remove the frontend source from the package (we will not distribute these)
        shutil.rmtree(frontend_path, ignore_errors=True)


class CleanCommand(Command):
    """Custom clean command to tidy up the project root."""
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    @staticmethod
    def run():
        shutil.rmtree(os.path.abspath(os.path.join(os.path.dirname(__file__), 'build')), ignore_errors=True)
        shutil.rmtree(os.path.abspath(os.path.join(os.path.dirname(__file__), 'dist')), ignore_errors=True)
        shutil.rmtree(os.path.abspath(os.path.join(os.path.dirname(__file__), '*.pyc')), ignore_errors=True)
        shutil.rmtree(os.path.abspath(os.path.join(os.path.dirname(__file__), 'unmanic.egg-info')), ignore_errors=True)
        [shutil.rmtree(f) for f in glob.glob(src_dir + "/**/__pycache__", recursive=True)]


class FullVersionCommand(Command):
    """Custom clean command to tidy up the project root."""
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    @staticmethod
    def run():
        print(versioninfo.full_version())


cmd_class = {
    'build_py':            BuildPyCommand,
    'write-build-version': WriteVersionCommand,
    'build-frontend':      BuildFrontendCommand,
    'clean':               CleanCommand,
    'fullversion':         FullVersionCommand,
}


def requirements():
    with open(os.path.abspath(os.path.join(os.path.dirname(__file__), 'requirements.txt'))) as f:
        return f.read().splitlines()


def requirements_dev():
    with open(os.path.abspath(os.path.join(os.path.dirname(__file__), 'requirements-dev.txt'))) as f:
        return f.read().splitlines()


setup(
    name=module_name,
    version=module_version,
    author=module_author,
    author_email=module_email,
    maintainer=module_author,
    maintainer_email=module_email,
    description=module_description,
    url=module_url,
    classifiers=module_classifiers,
    install_requires=requirements(),
    extras_require={
        'dev': requirements_dev()
    },
    packages=[src_dir],
    entry_points={
        'console_scripts': [
            '%s=%s.service:main' % (module_name, module_name)
        ]
    },
    cmdclass=cmd_class,
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.taskhandler.py

    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     08 May 2020, (12:22 PM)

    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved

           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:

           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.

           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""
import os
import queue
import threading
import time

from unmanic import config
from unmanic.libs import common, task
from unmanic.libs.unmodels.tasks import Tasks


class TaskHandler(threading.Thread):
    """
    TaskHandler

    The TaskHandler reads all items in the queues and passes them to the appropriate locations in the application.
        TODO:
            - All tasks are added to the database tasks table (no key to historical tasks.
                Row to be deleted once task is added to historical record)
            - Task Handler to monitor idle workers rather than idle workers looking for tasks in the TaskQueue object.
                - When a worker thread is idle, the TaskHandler needs to read a select query on the database and add that
                    item to the TaskQueue
            - Workers should request a job from the TaskHandler rather than reading the TaskQueue directly ??
            -
    """

    def __init__(self, data_queues, task_queue):
        super(TaskHandler, self).__init__(name='TaskHandler')
        self.settings = config.Config()
        self.data_queues = data_queues
        self.logger = data_queues["logging"].get_logger(self.name)
        self.task_queue = task_queue
        self.inotifytasks = data_queues["inotifytasks"]
        self.scheduledtasks = data_queues["scheduledtasks"]
        self.abort_flag = threading.Event()
        self.abort_flag.clear()
        # Remove all items from the task list to start with
        self.clear_tasks_on_startup()

    def _log(self, message, message2='', level="info"):
        message = common.format_message(message, message2)
        getattr(self.logger, level)(message)

    def stop(self):
        self.abort_flag.set()

    def run(self):
        self._log("Starting TaskHandler Monitor loop")
        while not self.abort_flag.is_set():
            self.process_scheduledtasks_queue()
            self.process_inotifytasks_queue()
            time.sleep(.2)

        self._log("Leaving TaskHandler Monitor loop...")

    def process_scheduledtasks_queue(self):
        while not self.abort_flag.is_set() and not self.scheduledtasks.empty():
            try:
                item = self.scheduledtasks.get_nowait()
                pathname = item['pathname']
                library_id = item['library_id']
                priority_score = item.get('priority_score', 0)
                if self.add_path_to_task_queue(pathname, library_id, priority_score=priority_score):
                    self._log("Adding file to task queue", pathname, level='info')
                else:
                    self._log("Skipping file as it is already in the queue", pathname, level='info')
            except queue.Empty:
                continue
            except Exception as e:
                self._log("Exception in processing scheduledtasks", str(e), level='exception')

    def process_inotifytasks_queue(self):
        while not self.abort_flag.is_set() and not self.inotifytasks.empty():
            try:
                item = self.inotifytasks.get_nowait()
                pathname = item['pathname']
                library_id = item['library_id']
                priority_score = item.get('priority_score', 0)
                # TODO: Ensure the file is not still being modified at this point.
                #  If it is still being modified here, it is ok to wait for that to finish (should not matter much)
                if self.add_path_to_task_queue(pathname, library_id, priority_score=priority_score):
                    self._log("Adding inotify job to queue", pathname, level='info')
                else:
                    self._log("Skipping inotify job already in the queue", pathname, level='info')
            except queue.Empty:
                continue
            except Exception as e:
                self._log("Exception in processing inotifytasks", str(e), level='exception')

    def clear_tasks_on_startup(self):
        query = Tasks.delete()
        if not self.settings.get_clear_pending_tasks_on_restart():
            # Exclude all pending tasks except for those that are remote tasks... They need to be removed
            query = query.where((Tasks.status != 'pending') | (Tasks.type == 'remote'))
        rows_deleted_count = query.execute()
        self._log("Deleted {} items from tasks list".format(rows_deleted_count), level='debug')

    @staticmethod
    def check_if_task_exists_matching_path(abspath):
        """
        Check if a task already exists matching the given path

        :param abspath:
        :return:
        """
        existing_task_query = Tasks.select().where((Tasks.abspath == abspath)).limit(1)
        if existing_task_query.count() > 0:
            return True
        return False

    def add_path_to_task_queue(self, pathname, library_id, priority_score=0):
        """
        Add the path to the task queue ensuring that the path is only added once

        :param pathname:
        :param library_id:
        :param priority_score:
        :return:
        """
        # Check if file exists in task queue based on it's absolute path
        abspath = os.path.abspath(pathname)
        if self.check_if_task_exists_matching_path(abspath):
            return False
        # Create the new task from the provide path
        new_task = self.create_task_from_path(pathname, library_id, priority_score=priority_score)
        if not new_task:
            return False
        return True

    def create_task_from_path(self, pathname, library_id, priority_score=0):
        """
        Generate a Task object from a pathname

        :param pathname:
        :param library_id:
        :param priority_score:
        :return:
        """
        abspath = os.path.abspath(pathname)
        # Create a new task
        new_task = task.Task()

        if not new_task.create_task_by_absolute_path(abspath, library_id=library_id, priority_score=priority_score):
            # If file exists in task queue already this will return false.
            # Do not carry on.
            return False

        return new_task
